"""G.711 and G.726 audio codec inspired utilities for time series processing.

This module provides:
1. G.711 companding (mu-law and A-law) for non-linear scaling
2. G.726 adaptive differential PCM for denoising/smoothing
3. Transformer classes (G711Scaler, G726Filter) for AutoTS integration

G.711 is used as a non-linear scaler: encode on transform, decode on inverse_transform.
G.726 is used as an adaptive filter to denoise while preserving structure.

All operations are fully vectorized for efficient processing of multiple time series.
"""

from __future__ import annotations

import random
from typing import Tuple, Union
import numpy as np
import pandas as pd

try:  # pragma: no cover - SciPy is expected but AutoTS has fallbacks.
    from scipy.signal import lfilter
except Exception:  # pragma: no cover
    lfilter = None

# Use consistent epsilon for numerical stability
EPS = 1e-12
EPSILON = np.finfo(float).eps

# ============================================================================
# G.711 Companding (Mu-law and A-law)
# ============================================================================


def _prep_input(values: Union[np.ndarray, pd.DataFrame]) -> Tuple[np.ndarray, bool]:
    """Return a float64 array with shape (n_obs, n_series) and squeeze flag."""
    if isinstance(values, pd.DataFrame):
        arr = values.to_numpy(dtype=float, copy=True)
    else:
        arr = np.asarray(values, dtype=float)
    squeezed = False
    if arr.ndim == 1:
        arr = arr[:, None]
        squeezed = True
    return arr.copy(), squeezed


def _robust_center_scale(
    arr: np.ndarray,
    center: str = "median",
    scale_method: str = "mad",
    scale_factor: float = 3.0,
    min_scale: float = 1e-6,
) -> Tuple[np.ndarray, np.ndarray]:
    """Compute per-series center and scale using robust statistics.

    Parameters
    ----------
    arr : ndarray of shape (n_obs, n_series)
    center : {'median','mean'}
    scale_method : {'mad','std','maxabs','percentile'}
    scale_factor : float multiplier applied to scale estimate
    min_scale : minimum scale to avoid division by zero
    """
    if center == "mean":
        c = np.nanmean(arr, axis=0, keepdims=True)
    else:
        c = np.nanmedian(arr, axis=0, keepdims=True)

    x = arr - c
    sm = scale_method.lower()
    if sm == "std":
        s = np.nanstd(x, axis=0, ddof=0, keepdims=True)
    elif sm == "maxabs":
        s = np.nanmax(np.abs(x), axis=0, keepdims=True)
    elif sm == "percentile":
        # Use 90th percentile of absolute deviations for better sparse data handling
        s = np.nanpercentile(np.abs(x), 90, axis=0, keepdims=True)
    else:  # 'mad' (scaled to be comparable to std under normality)
        med = np.nanmedian(np.abs(x), axis=0, keepdims=True)
        s = 1.4826 * med

        # Fallback to percentile if MAD is suspiciously small (indicates sparse data)
        # This helps with intermittent series where median deviation is near zero
        percentile_scale = np.nanpercentile(np.abs(x), 90, axis=0, keepdims=True)
        too_small = s < (percentile_scale * 0.1)  # MAD < 10% of p90
        if np.any(too_small):
            s = np.where(too_small, percentile_scale, s)

    s = np.maximum(s * float(scale_factor), float(min_scale))
    return c, s


def _mu_law_compress(x: np.ndarray, mu: float = 255.0) -> np.ndarray:
    """Mu-law companding for normalized input in [-1, 1]."""
    mu = float(mu)
    x = np.clip(x, -1.0, 1.0)
    return np.sign(x) * np.log1p(mu * np.abs(x)) / np.log1p(mu)


def _mu_law_expand(y: np.ndarray, mu: float = 255.0) -> np.ndarray:
    """Inverse mu-law for y in [-1, 1]."""
    mu = float(mu)
    y = np.clip(y, -1.0, 1.0)
    return np.sign(y) * (np.expm1(np.abs(y) * np.log1p(mu)) / mu)


def _a_law_compress(x: np.ndarray, A: float = 87.6) -> np.ndarray:
    """A-law companding for normalized input in [-1, 1]."""
    A = float(A)
    x = np.clip(x, -1.0, 1.0)
    ax = np.abs(x)
    k = 1.0 / (1.0 + np.log(A))
    # piecewise without explicit Python loops
    small = ax <= (1.0 / A)
    y = np.empty_like(x)
    y[small] = np.sign(x[small]) * (A * ax[small]) * k
    y[~small] = np.sign(x[~small]) * (1.0 + np.log(A * ax[~small])) * k
    return y


def _a_law_expand(y: np.ndarray, A: float = 87.6) -> np.ndarray:
    """Inverse A-law for y in [-1, 1]."""
    A = float(A)
    y = np.clip(y, -1.0, 1.0)
    ay = np.abs(y)
    k = 1.0 + np.log(A)
    thresh = 1.0 / k
    x = np.empty_like(y)
    small = ay <= thresh
    x[small] = np.sign(y[small]) * (ay[small] * k) / A
    # exp component
    x[~small] = np.sign(y[~small]) * (np.exp(ay[~small] * k - 1.0) / A)
    return x


def g711_encode(
    values: Union[np.ndarray, pd.DataFrame],
    mode: str = "mu",
    mu: float = 255.0,
    A: float = 87.6,
    center: str = "median",
    scale_method: str = "mad",
    scale_factor: float = 3.0,
    min_scale: float = 1e-6,
    clip: bool = True,
    zero_offset: float = 0.0,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Encode values via G.711-style companding after robust normalization.

    Returns encoded array and the center/scale used so that decoding can
    recover the original feature space.

    Parameters
    ----------
    zero_offset : float, optional
        Small constant added before transformation to handle exact zeros.
        Useful for sparse/intermittent data. Default 0.0 (no offset).
    """
    arr, squeezed = _prep_input(values)
    if arr.size == 0:
        # Return empty arrays with appropriate shapes
        return arr, np.array([]), np.array([])

    # Apply zero offset if specified
    if zero_offset != 0.0:
        arr = arr + zero_offset

    c, s = _robust_center_scale(
        arr,
        center=center,
        scale_method=scale_method,
        scale_factor=scale_factor,
        min_scale=min_scale,
    )
    x = (arr - c) / (s + EPS)
    if clip:
        x = np.clip(x, -1.0, 1.0)

    m = mode.lower()
    if m in ("mu", "mulaw", "u", "ulaw"):
        y = _mu_law_compress(x, mu=mu)
    elif m in ("a", "alaw"):
        y = _a_law_compress(x, A=A)
    else:
        raise ValueError(f"Unknown G711 mode: {mode}")

    return y if not squeezed else y[:, 0], c, s


def g711_decode(
    encoded: Union[np.ndarray, pd.DataFrame],
    center: np.ndarray,
    scale: np.ndarray,
    mode: str = "mu",
    mu: float = 255.0,
    A: float = 87.6,
    zero_offset: float = 0.0,
) -> np.ndarray:
    """Decode from G.711 companded space back to original feature space.

    Parameters
    ----------
    zero_offset : float, optional
        Must match the offset used in encoding. Subtracted after decoding.
    """
    arr, squeezed = _prep_input(encoded)
    if arr.size == 0:
        return arr

    m = mode.lower()
    if m in ("mu", "mulaw", "u", "ulaw"):
        x = _mu_law_expand(arr, mu=mu)
    elif m in ("a", "alaw"):
        x = _a_law_expand(arr, A=A)
    else:
        raise ValueError(f"Unknown G711 mode: {mode}")

    out = x * (scale + EPS) + center

    # Remove zero offset if it was applied
    if zero_offset != 0.0:
        out = out - zero_offset

    return out if not squeezed else out[:, 0]


# ============================================================================
# G.726 Adaptive Differential PCM
# ============================================================================

# Non-uniform quantizer decision levels and reconstruction values
# Optimized for daily time series data (4-bit variant, 16 levels)
# Based on G.726 32kbps tables but tuned for daily/hourly patterns
QUANTIZER_DECISION_LEVELS = np.array(
    [-7.5, -5.5, -4.0, -3.0, -2.0, -1.3, -0.7, 0.0, 0.7, 1.3, 2.0, 3.0, 4.0, 5.5, 7.5]
)

QUANTIZER_RECONSTRUCTION = np.array(
    [
        -8.5,
        -6.5,
        -4.75,
        -3.5,
        -2.5,
        -1.65,
        -1.0,
        -0.35,
        0.35,
        1.0,
        1.65,
        2.5,
        3.5,
        4.75,
        6.5,
        8.5,
    ]
)

# Scale factor adaptation multipliers (based on G.726)
# Index by quantizer output (0-15 for 4-bit)
SCALE_FACTOR_MULTIPLIERS = np.array(
    [
        0.60,
        0.93,
        0.93,
        0.93,
        1.20,
        1.20,
        1.20,
        1.20,
        1.20,
        1.20,
        1.20,
        1.20,
        0.93,
        0.93,
        0.93,
        0.60,
    ]
)

# Adaptation speed control
FAST_SCALE_ALPHA = 0.875  # Fast adaptation for scale factor
SLOW_SCALE_ALPHA = 0.985  # Slow adaptation for scale factor
PREDICTOR_LEAK = 0.9999  # Coefficient leakage factor


def _adaptive_predictor_step(
    a1: np.ndarray,
    a2: np.ndarray,
    b: np.ndarray,
    sez_history: np.ndarray,
    dq: np.ndarray,
    leak: float = PREDICTOR_LEAK,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Update 2-pole/6-zero predictor coefficients and state.

    Parameters
    ----------
    a1, a2 : ndarray
        Adaptive pole coefficients (shape: n_series)
    b : ndarray
        Adaptive zero coefficients (shape: 6, n_series)
    sez_history : ndarray
        Quantized difference signal history (shape: 6, n_series)
    dq : ndarray
        Quantized difference signal (current sample, shape: n_series)
    leak : float
        Coefficient leakage factor to prevent unbounded growth

    Returns
    -------
    a1_new, a2_new, b_new, sez_new : tuple of ndarrays
        Updated coefficients and shifted state
    """
    # Shift quantized difference history
    sez_new = np.roll(sez_history, 1, axis=0)
    sez_new[0] = dq

    # Gradient adaptation using sign-based LMS (more stable than full LMS)
    # Very small step sizes for stability
    sgn_dq = np.sign(dq)

    # Update pole coefficients with strong leakage
    mu_a = 0.001  # Very small step size
    a1_new = leak * a1 + mu_a * sgn_dq * np.sign(sez_history[0])
    a2_new = leak * a2 + mu_a * sgn_dq * np.sign(sez_history[1])

    # Strong clamping for stability (poles must be inside unit circle)
    a1_new = np.clip(a1_new, -0.7, 0.7)
    a2_new = np.clip(a2_new, -0.7, 0.7)

    # Update zero coefficients
    mu_b = 0.0005  # Even smaller for zeros
    b_new = np.copy(b)
    for i in range(6):
        b_new[i] = leak * b[i] + mu_b * sgn_dq * np.sign(sez_history[i])
        b_new[i] = np.clip(b_new[i], -1.0, 1.0)

    return a1_new, a2_new, b_new, sez_new


def _compute_prediction(
    a1: np.ndarray,
    a2: np.ndarray,
    b: np.ndarray,
    sr_history: np.ndarray,
    sez_history: np.ndarray,
) -> np.ndarray:
    """Compute adaptive prediction from pole-zero model.

    Parameters
    ----------
    a1, a2 : ndarray
        Pole coefficients (shape: n_series)
    b : ndarray
        Zero coefficients (shape: 6, n_series)
    sr_history : ndarray
        Past reconstructed samples (shape: 2, n_series)
    sez_history : ndarray
        Past quantized differences (shape: 6, n_series)

    Returns
    -------
    ndarray
        Predicted value (shape: n_series)
    """
    # Clip inputs to prevent numerical issues
    sr_history = np.clip(sr_history, -1e6, 1e6)
    sez_history = np.clip(sez_history, -1e6, 1e6)

    # Pole contribution (2nd order recursive on reconstructed signal)
    pole_pred = a1 * sr_history[0] + a2 * sr_history[1]

    # Zero contribution (6-tap FIR on quantized differences)
    zero_pred = (b * sez_history).sum(axis=0)

    # Combined prediction
    prediction = pole_pred + zero_pred

    # Clip to prevent overflow
    prediction = np.clip(prediction, -1e6, 1e6)

    return prediction


def _quantize_uniform(residual: np.ndarray, scale: np.ndarray, bits: int) -> np.ndarray:
    """Uniform quantization (original method)."""
    max_level = (1 << (bits - 1)) - 1
    min_level = -(1 << (bits - 1))
    return np.clip(np.round(residual / (scale + EPSILON)), min_level, max_level)


def _quantize_nonuniform(
    residual: np.ndarray, scale: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Non-uniform quantization using decision levels.

    Returns
    -------
    indices : ndarray
        Quantizer indices (0-15 for 4-bit)
    reconstructed : ndarray
        Reconstructed values
    """
    # Normalize by scale factor
    normalized = residual / (scale + EPSILON)

    # Handle NaN/Inf (replace with zero) and clamp to sane range
    normalized = np.nan_to_num(normalized, nan=0.0, posinf=10.0, neginf=-10.0)
    normalized = np.clip(normalized, -15.0, 15.0)

    # Find quantizer bin using decision levels
    indices = np.searchsorted(QUANTIZER_DECISION_LEVELS, normalized)

    # Get reconstruction values and scale back
    reconstructed = QUANTIZER_RECONSTRUCTION[indices] * scale

    return indices, reconstructed


def _update_scale_factor(
    scale: np.ndarray,
    quant_index: np.ndarray,
    fast_mode: np.ndarray,
) -> np.ndarray:
    """Update scale factor using G.726 adaptation logic.

    Parameters
    ----------
    scale : ndarray
        Current scale factor (shape: n_series)
    quant_index : ndarray
        Quantizer output indices (shape: n_series)
    fast_mode : ndarray
        Boolean array indicating fast adaptation mode (shape: n_series)

    Returns
    -------
    ndarray
        Updated scale factor
    """
    # Get multiplier from lookup table
    multiplier = SCALE_FACTOR_MULTIPLIERS[quant_index.astype(int)]

    # Choose adaptation speed
    alpha = np.where(fast_mode, FAST_SCALE_ALPHA, SLOW_SCALE_ALPHA)

    # Update: scale = alpha * scale + (1 - alpha) * (multiplier * scale)
    # Simplified: scale = scale * (alpha + (1 - alpha) * multiplier)
    scale_new = scale * (alpha + (1.0 - alpha) * multiplier)

    # Bound scale factor (prevent too small or too large)
    scale_new = np.clip(scale_new, 0.001, 100.0)

    return scale_new


def g726_adpcm_filter(
    values: Union[np.ndarray, pd.DataFrame],
    quant_bits: int = 4,
    adaptation_rate: float = 0.96,
    prediction_alpha: float = 0.92,
    floor_step: float = 0.01,
    dynamic_range: float = 1.5,
    blend: float = 0.15,
    noise_gate: float = 0.0,
    quantizer: str = "uniform",
    use_adaptive_predictor: bool = True,
    predictor_leak: float = PREDICTOR_LEAK,
) -> np.ndarray:
    """Apply the adaptive encode/decode cycle to an array of time series.

    Parameters
    ----------
    values : array-like or DataFrame
        Shape (observations, series). Each column is processed independently,
        but operations stay vectorized across series.
    quant_bits : int, default 4
        Number of bits used for the quantizer (only for uniform mode).
    adaptation_rate : float, default 0.96
        Only used when use_adaptive_predictor=False (legacy EMA mode).
    prediction_alpha : float, default 0.92
        Only used when use_adaptive_predictor=False (legacy EMA mode).
    floor_step : float, default 0.01
        Initial scale factor (minimum allowed value).
    dynamic_range : float, default 1.5
        Initial scale multiplier (legacy mode only).
    blend : float, default 0.15
        Optional blending factor with the predictor baseline. Must be in [0, 1].
    noise_gate : float, default 0.0
        Additional soft threshold applied to residual magnitudes.
    quantizer : str, default "uniform"
        Quantization method: "uniform" for simple rounding, "nonuniform" for
        G.726-style decision levels optimized for time series.
    use_adaptive_predictor : bool, default True
        Use 2-pole/6-zero adaptive predictor. If False, falls back to EMA.
    predictor_leak : float, default 0.9999
        Leakage factor for predictor coefficients to prevent unbounded growth.

    Returns
    -------
    numpy.ndarray
        Filtered data with the same shape as ``values``.
    """

    arr, squeezed = _prep_input(values)
    if arr.size == 0 or arr.shape[0] <= 1:
        return arr if not squeezed else arr[:, 0]

    n_obs, n_series = arr.shape
    blend = float(np.clip(blend, 0.0, 1.0))

    if not use_adaptive_predictor:
        # Legacy mode: simple EMA predictor (backward compatibility)
        return _legacy_filter(
            arr,
            squeezed,
            quant_bits,
            adaptation_rate,
            prediction_alpha,
            floor_step,
            dynamic_range,
            blend,
            noise_gate,
        )

    # Initialize adaptive predictor state (vectorized across series)
    a1 = np.zeros(n_series)  # 1st pole coefficient
    a2 = np.zeros(n_series)  # 2nd pole coefficient
    b = np.zeros((6, n_series))  # 6 zero coefficients

    # History buffers
    sr_history = np.zeros((2, n_series))  # Past reconstructed samples (for poles)
    sez_history = np.zeros((6, n_series))  # Past quantized differences (for zeros)

    # Scale factor initialization - estimate from first few samples
    if n_obs > 5:
        initial_scale = np.maximum(np.std(arr[:10], axis=0), floor_step)
    else:
        initial_scale = np.full(n_series, floor_step, dtype=float)
    scale = np.asarray(initial_scale, dtype=float).copy()

    # Track fast/slow adaptation mode
    fast_mode = np.ones(n_series, dtype=bool)  # Start in fast mode
    transition_count = np.zeros(n_series, dtype=int)

    # Output array
    filtered = np.empty_like(arr)
    filtered[0] = arr[0]  # First sample passes through
    sr_history[0] = arr[0]  # Initialize with first sample level

    # Synchronous coding: iterate over time, vectorize across series
    for t in range(1, n_obs):
        # Compute prediction from adaptive model
        prediction = _compute_prediction(a1, a2, b, sr_history, sez_history)

        # Residual (difference from prediction)
        residual = arr[t] - arr[t - 1] - prediction

        # Optional noise gate
        if noise_gate > 0.0:
            gate = np.maximum(0.0, 1.0 - (noise_gate / (np.abs(residual) + noise_gate)))
            residual = residual * gate

        # Quantization
        if quantizer == "nonuniform":
            quant_idx, reconstructed_diff = _quantize_nonuniform(residual, scale)
        else:
            # Uniform quantization (legacy mode)
            # Note: Scale adaptation LUT is optimized for 4-bit; other bit depths
            # will have approximate adaptation behavior
            bits = int(np.clip(np.round(quant_bits), 2, 6))
            encoded = _quantize_uniform(residual, scale, bits)
            reconstructed_diff = encoded * scale
            # Map to indices for scale adaptation (approximate for non-4-bit)
            quant_idx = (encoded + (1 << (bits - 1))).astype(int)
            quant_idx = np.clip(quant_idx, 0, 15)

        # Reconstruct signal (integrate differences)
        reconstructed = arr[t - 1] + prediction + reconstructed_diff

        # Clip to prevent overflow
        reconstructed = np.clip(reconstructed, -1e8, 1e8)
        filtered[t] = reconstructed

        # Update scale factor with adaptive speed
        scale = _update_scale_factor(scale, quant_idx, fast_mode)

        # Update fast/slow mode based on stability
        # Switch to slow mode if signal is stable (center quantizer indices)
        stable = (quant_idx >= 6) & (quant_idx <= 9)
        transition_count = np.where(stable, transition_count + 1, 0)
        fast_mode = transition_count < 5  # Require 5 stable samples to switch

        # Update predictor coefficients and state (uses quantized differences)
        a1, a2, b, sez_history = _adaptive_predictor_step(
            a1, a2, b, sez_history, reconstructed_diff, predictor_leak
        )

        # Update reconstructed signal history for poles
        sr_history = np.roll(sr_history, 1, axis=0)
        sr_history[0] = reconstructed

    # Optional blending with slow-moving baseline
    if blend > 0.0:
        # Compute simple baseline for blending using prediction_alpha
        if lfilter is not None:
            alpha = float(np.clip(prediction_alpha, 0.0, 0.9999))
            baseline = lfilter([1.0 - alpha], [1.0, -alpha], arr, axis=0)
        else:
            baseline = arr  # Skip blending if no scipy
        filtered = (1.0 - blend) * filtered + blend * baseline

    return filtered if not squeezed else filtered[:, 0]


def _legacy_filter(
    arr: np.ndarray,
    squeezed: bool,
    quant_bits: int,
    adaptation_rate: float,
    prediction_alpha: float,
    floor_step: float,
    dynamic_range: float,
    blend: float,
    noise_gate: float,
) -> np.ndarray:
    """Legacy EMA-based filter (for backward compatibility)."""
    bits = int(np.clip(np.round(quant_bits), 2, 6))

    # Simple EMA predictor (non-synchronous, uses original signal)
    if lfilter is not None:
        alpha = float(np.clip(prediction_alpha, 0.0, 0.9999))
        baseline = lfilter([1.0 - alpha], [1.0, -alpha], arr, axis=0)
    else:
        baseline = arr * 0.5  # Fallback

    baseline[:1, :] = arr[:1, :]
    residual = arr - baseline

    if noise_gate > 0.0:
        gate = np.maximum(0.0, 1.0 - (noise_gate / (np.abs(residual) + noise_gate)))
        residual = residual * gate

    # Simple scale adaptation
    if lfilter is not None:
        alpha = float(np.clip(adaptation_rate, 0.0, 0.9999))
        scale = lfilter([1.0 - alpha], [1.0, -alpha], np.abs(residual), axis=0)
    else:
        scale = np.abs(residual)

    scale = floor_step + dynamic_range * scale
    scale = np.maximum(scale, floor_step)

    # Quantize
    encoded = _quantize_uniform(residual, scale, bits)
    decoded = encoded * scale
    filtered = baseline + decoded
    filtered[:1, :] = arr[:1, :]

    if blend > 0.0:
        filtered = (1.0 - blend) * filtered + blend * baseline

    return filtered if not squeezed else filtered[:, 0]


# ============================================================================
# Transformer Classes for AutoTS Integration
# ============================================================================


class EmptyTransformer(object):
    """Base transformer returning raw data."""

    def __init__(self, name: str = "EmptyTransformer", **kwargs):
        self.name = name

    def _fit(self, df):
        """Learn behavior of data to change.

        Args:
            df (pandas.DataFrame): input dataframe
        """
        return df

    def fit(self, df):
        """Learn behavior of data to change.

        Args:
            df (pandas.DataFrame): input dataframe
        """
        self._fit(df)
        return self

    def transform(self, df):
        """Transform data.

        Args:
            df (pandas.DataFrame): input dataframe
        """
        return df

    def inverse_transform(self, df, trans_method: str = "forecast", adjustment=None):
        """Inverse transform data.

        Args:
            df (pandas.DataFrame): input dataframe
        """
        return df

    def fit_transform(self, df):
        """Fit and Transform data.

        Args:
            df (pandas.DataFrame): input dataframe
        """
        self.fit(df)
        return self.transform(df)


class G711Scaler(EmptyTransformer):
    """G.711-style non-linear scaler (mu-law or A-law).

    Encodes (compands) on transform and decodes (expands to original space)
    on inverse_transform. Optimized for daily/hourly data by using robust
    per-series centering and scaling to normalize into [-1, 1] prior to
    companding.

    Args:
        mode: 'mu' for mu-law or 'a' for A-law companding
        mu: mu-law parameter (default 255.0, try 100-512 range)
        A: A-law parameter (default 87.6, try 50-100 range)
        center: 'median' or 'mean' for centering
        scale_method: 'mad', 'std', 'maxabs', or 'percentile' for scale estimation
        scale_factor: Multiplier for scale (default 3.0, lower=more compression, higher=better reconstruction)
        min_scale: Minimum scale to prevent division by zero
        clip: Whether to clip normalized values to [-1, 1]
        zero_offset: Small constant added to handle exact zeros (e.g., 1e-6 for sparse data)
        fill_method: Method to handle NaN values
        on_transform: Apply encoding on transform
        on_inverse: Apply decoding on inverse_transform
        bounds_only: Only apply during bounds adjustment

    Note:
        For sparse/intermittent data with many zeros, consider setting zero_offset
        or using alternative transformers like ReplaceConstant or CenterSplit.
    """

    def __init__(
        self,
        mode: str = "mu",
        mu: float = 100.0,
        A: float = 87.6,
        center: str = "median",
        scale_method: str = "mad",
        scale_factor: float = 3.0,
        min_scale: float = 1e-6,
        clip: bool = True,
        zero_offset: float = 0.0,
        fill_method: str = "interpolate",
        on_transform: bool = True,
        on_inverse: bool = True,
        bounds_only: bool = False,
        **kwargs,
    ):
        super().__init__(name="G711Scaler")
        self.mode = mode
        self.mu = float(mu)
        self.A = float(A)
        self.center = center
        self.scale_method = scale_method
        self.scale_factor = float(scale_factor)
        self.min_scale = float(min_scale)
        self.clip = bool(clip)
        self.zero_offset = float(zero_offset)
        self.fill_method = fill_method
        self.on_transform = on_transform
        self.on_inverse = on_inverse
        self.bounds_only = bounds_only
        self._center = None
        self._scale = None
        self.adjustment = None

    def fit(self, df):
        self.columns = df.columns
        frame = self._prepare_frame(df)
        # use g711_encode to compute and store center/scale without consuming time
        _, c, s = g711_encode(
            frame.to_numpy(),
            mode=self.mode,
            mu=self.mu,
            A=self.A,
            center=self.center,
            scale_method=self.scale_method,
            scale_factor=self.scale_factor,
            min_scale=self.min_scale,
            clip=self.clip,
            zero_offset=self.zero_offset,
        )
        # store with 2D shape for broadcasting
        if c.size == 0:
            self._center = np.zeros((1, 0))
            self._scale = np.ones((1, 0))
        else:
            self._center = c
            self._scale = s
        return self

    def _prepare_frame(self, df: pd.DataFrame) -> pd.DataFrame:
        frame = df.apply(pd.to_numeric, errors="coerce")
        if self.fill_method == "interpolate":
            frame = frame.interpolate(limit_direction="both")
        elif self.fill_method == "ffill":
            frame = frame.ffill().bfill()
        elif self.fill_method == "bfill":
            frame = frame.bfill().ffill()
        elif self.fill_method == "median":
            frame = frame.fillna(frame.median())
        else:
            frame = frame.fillna(0.0)
        return frame.fillna(0.0)

    def transform(self, df):
        if not self.on_transform:
            return df
        if df.empty:
            return df
        frame = self._prepare_frame(df)
        arr = frame.to_numpy(dtype=float)

        # Apply zero offset if specified
        if self.zero_offset != 0.0:
            arr = arr + self.zero_offset

        # Normalize using stored center/scale and encode (consistent EPS)
        x = (arr - self._center) / (self._scale + 1e-12)
        if self.clip:
            x = np.clip(x, -1.0, 1.0)
        m = self.mode.lower()
        if m in ("mu", "mulaw", "u", "ulaw"):
            encoded = _mu_law_compress(x, mu=self.mu)
        elif m in ("a", "alaw"):
            encoded = _a_law_compress(x, A=self.A)
        else:
            raise ValueError(f"Unknown G711 mode: {self.mode}")
        return pd.DataFrame(encoded, index=df.index, columns=df.columns)

    def inverse_transform(self, df, trans_method: str = "forecast", adjustment=None):
        if not self.on_inverse:
            return df
        if self.bounds_only and adjustment is None:
            self.adjustment = True
            return df
        if df.empty:
            return df
        arr = df.apply(pd.to_numeric, errors="coerce").fillna(0.0).to_numpy(dtype=float)
        decoded = g711_decode(
            arr,
            center=self._center,
            scale=self._scale,
            mode=self.mode,
            mu=self.mu,
            A=self.A,
            zero_offset=self.zero_offset,
        )
        return pd.DataFrame(decoded, index=df.index, columns=self.columns)

    def fit_transform(self, df):
        self.fit(df)
        return self.transform(df)

    @staticmethod
    def get_new_params(method: str = "random"):
        mode = random.choices(["mu", "a"], [0.7, 0.3])[0]
        return {
            "mode": mode,
            # Reduced mu values to lessen non-linear dampening
            # Favor lower values (50-150) for better forecast amplitude preservation
            "mu": random.choices(
                [50.0, 75.0, 100.0, 150.0, 200.0, 255.0],
                [0.25, 0.25, 0.2, 0.15, 0.1, 0.05],
            )[0]
            if mode == "mu"
            else 100.0,
            "A": random.choice([50.0, 75.0, 87.6, 100.0]) if mode == "a" else 87.6,
            "center": random.choices(["median", "mean"], [0.8, 0.2])[0],
            "scale_method": random.choices(
                ["mad", "std", "maxabs", "percentile"], [0.5, 0.2, 0.1, 0.2]
            )[0],
            "scale_factor": random.uniform(1.5, 3.5),
            "min_scale": 1e-6,
            "clip": random.choice([True, True, False]),
            "zero_offset": random.choice([0.0, 0.0, 0.0, 1e-6, 1e-8]),
            "fill_method": random.choices(
                ["interpolate", "ffill", "median"], [0.6, 0.3, 0.1]
            )[0],
            "on_transform": True,
            "on_inverse": True,
            "bounds_only": False,
        }


class G726Filter(EmptyTransformer):
    """Adaptive differential PCM smoothing inspired by the G.726 codec.

    Implements proper G.726-style adaptive filtering with:
    - 2-pole/6-zero adaptive predictor with decision-directed updates (not unlike an ARMA (2,6) model)
    - Non-uniform quantization optimized for time series
    - Fast/slow scale factor adaptation with speed switching
    """

    def __init__(
        self,
        quant_bits: int = 4,
        adaptation_rate: float = 0.96,
        prediction_alpha: float = 0.92,
        floor_step: float = 0.01,
        dynamic_range: float = 1.5,
        blend: float = 0.15,
        noise_gate: float = 0.0,
        fill_method: str = "interpolate",
        on_transform: bool = True,
        on_inverse: bool = False,
        bounds_only: bool = False,
        quantizer: str = "uniform",
        use_adaptive_predictor: bool = True,
        predictor_leak: float = 0.9999,
        **kwargs,
    ):
        super().__init__(name="G726Filter")
        self.quant_bits = quant_bits
        self.adaptation_rate = adaptation_rate
        self.prediction_alpha = prediction_alpha
        self.floor_step = floor_step
        self.dynamic_range = dynamic_range
        self.blend = blend
        self.noise_gate = noise_gate
        self.fill_method = fill_method
        self.on_transform = on_transform
        self.on_inverse = on_inverse
        self.bounds_only = bounds_only
        self.quantizer = quantizer
        self.use_adaptive_predictor = use_adaptive_predictor
        self.predictor_leak = predictor_leak
        self.adjustment = None

    def fit(self, df):
        self.columns = df.columns
        return self

    def _prepare_frame(self, df: pd.DataFrame) -> pd.DataFrame:
        frame = df.apply(pd.to_numeric, errors="coerce")
        if self.fill_method == "interpolate":
            frame = frame.interpolate(limit_direction="both")
        elif self.fill_method == "ffill":
            frame = frame.ffill().bfill()
        elif self.fill_method == "bfill":
            frame = frame.bfill().ffill()
        elif self.fill_method == "median":
            frame = frame.fillna(frame.median())
        else:
            frame = frame.fillna(0.0)
        return frame.fillna(0.0)

    def _filter(self, df: pd.DataFrame) -> pd.DataFrame:
        if df.empty:
            return df
        prepared = self._prepare_frame(df)
        filtered = g726_adpcm_filter(
            prepared.to_numpy(),
            quant_bits=self.quant_bits,
            adaptation_rate=self.adaptation_rate,
            prediction_alpha=self.prediction_alpha,
            floor_step=self.floor_step,
            dynamic_range=self.dynamic_range,
            blend=self.blend,
            noise_gate=self.noise_gate,
            quantizer=self.quantizer,
            use_adaptive_predictor=self.use_adaptive_predictor,
            predictor_leak=self.predictor_leak,
        )
        return pd.DataFrame(filtered, index=df.index, columns=df.columns)

    def transform(self, df):
        if self.on_transform:
            return self._filter(df)
        return df

    def inverse_transform(self, df, trans_method: str = "forecast", adjustment=None):
        if self.on_inverse:
            if not self.bounds_only or (self.bounds_only and adjustment is not None):
                return self._filter(df)
            self.adjustment = True
        return df

    def fit_transform(self, df):
        self.fit(df)
        return self.transform(df)

    @staticmethod
    def get_new_params(method: str = "random"):
        selection = random.choices([True, False], [0.75, 0.25])[0]
        quant_bits = random.choices([3, 4, 5], [0.2, 0.6, 0.2])[0]
        use_adaptive = random.choices([True, False], [0.8, 0.2])[0]
        params = {
            "quant_bits": quant_bits,
            "adaptation_rate": random.uniform(0.9, 0.99),
            "prediction_alpha": random.uniform(0.7, 0.97),
            "floor_step": random.uniform(0.002, 0.05),
            "dynamic_range": random.uniform(0.5, 2.5),
            "blend": random.uniform(0.0, 0.3),
            "noise_gate": random.choice([0.0, random.uniform(0.0, 0.1)]),
            "fill_method": random.choices(
                ["interpolate", "ffill", "median"], [0.6, 0.3, 0.1]
            )[0],
            "on_transform": selection,
            "on_inverse": not selection,
            "quantizer": random.choices(["uniform", "nonuniform"], [0.4, 0.6])[0],
            "use_adaptive_predictor": use_adaptive,
            "predictor_leak": random.uniform(0.9995, 0.99995)
            if use_adaptive
            else 0.9999,
        }
        if not selection:
            params["bounds_only"] = random.choices([True, False], [0.2, 0.8])[0]
        else:
            params["bounds_only"] = False
        return params


__all__ = [
    "g711_encode",
    "g711_decode",
    "g726_adpcm_filter",
    "G711Scaler",
    "G726Filter",
]



# -*- coding: utf-8 -*-
"""Overall testing."""
import unittest
import json
import time
import timeit
import tempfile
import os
from unittest.mock import patch

import numpy as np
import pandas as pd
from autots.datasets import load_daily, load_monthly, load_artificial, load_sine
from autots import AutoTS, model_forecast, ModelPrediction, GeneralTransformer
from autots.tools.regressor import fake_regressor
from autots.evaluator.auto_model import (
    ModelMonster,
    TemplateEvalObject,
    reset_interrupt_tracking,
    _register_interrupt_press,
    _normalize_interrupt_mode,
)
from autots.models.ensemble import full_ensemble_test_list
from autots.models.model_list import default as default_model_list
from autots.models.model_list import all_models
from autots.evaluator.benchmark import Benchmark
from autots.templates.general import general_template
from autots.tools.cpu_count import cpu_count, set_n_jobs


class AutoTSTest(unittest.TestCase):
    def test_autots(self):
        print("Starting AutoTS class tests")
        forecast_length = 8
        long = False
        df = load_daily(long=long)
        n_jobs = 'auto'
        verbose = 0
        validation_method = "backwards"
        generations = 1
        num_validations = 2
        models_to_validate = 0.25  # must be a decimal percent for this test

        model_list = [
            'ConstantNaive',
            'LastValueNaive',
            'AverageValueNaive',
            'SeasonalNaive',
            'DatepartRegression',
        ]

        transformer_list = "all"  # ["SinTrend", "MinMaxScaler"]
        transformer_max_depth = 3

        metric_weighting = {
            'smape_weighting': 3,
            'mae_weighting': 1,
            'rmse_weighting': 1,
            'containment_weighting': 0,
            'runtime_weighting': 0,
            'spl_weighting': 1,
            'contour_weighting': 1,
        }
        # need to remove "profile" as it messes with tests below
        ensemble = [x for x in full_ensemble_test_list if "profile" not in x]

        model = AutoTS(
            forecast_length=forecast_length,
            frequency='infer',
            prediction_interval=0.9,
            ensemble=ensemble,
            constraint=None,
            max_generations=generations,
            num_validations=num_validations,
            validation_method=validation_method,
            model_list=model_list,
            transformer_list=transformer_list,
            transformer_max_depth=transformer_max_depth,
            initial_template='General+Random',
            metric_weighting=metric_weighting,
            models_to_validate=models_to_validate,
            max_per_model_class=None,
            model_interrupt="end_generation",
            no_negatives=True,
            subset=100,
            n_jobs=n_jobs,
            drop_most_recent=1,
            verbose=verbose,
        )
        future_regressor_train2d, future_regressor_forecast2d = fake_regressor(
            df,
            dimensions=4,
            forecast_length=forecast_length,
            date_col='datetime' if long else None,
            value_col='value' if long else None,
            id_col='series_id' if long else None,
            drop_most_recent=model.drop_most_recent,
            aggfunc=model.aggfunc,
            verbose=model.verbose,
        )
        model = model.fit(
            df,
            future_regressor=future_regressor_train2d,
            date_col='datetime' if long else None,
            value_col='value' if long else None,
            id_col='series_id' if long else None,
        )
        # first test multiple prediction intervals
        prediction = model.predict(
            future_regressor=future_regressor_forecast2d,
            prediction_interval=[0.6, 0.9],
            verbose=0,
        )
        prediction = model.predict(
            future_regressor=future_regressor_forecast2d, verbose=0
        )
        long_form = prediction.long_form_results()
        forecasts_df = prediction.forecast
        initial_results = model.results()
        validation_results = model.results("validation")
        back_forecast = model.back_forecast(n_splits=2, verbose=0).forecast
        # validated_count = (validation_results['Runs'] == (num_validations + 1)).sum()

        # so these account for DROP MOST RECENT = 1
        expected_idx = pd.date_range(
            start=df.index[-2], periods=forecast_length + 1, freq='D'
        )[1:]
        expected_val1 = pd.date_range(
            end=df.index[-(forecast_length + 2)], periods=forecast_length, freq='D'
        )
        expected_val2 = pd.date_range(
            end=df.index[-(forecast_length * 2 + 2)], periods=forecast_length, freq='D'
        )

        template_dict = json.loads(model.best_model['ModelParameters'].iloc[0])
        best_model_result = validation_results[
            validation_results['ID'] == model.best_model['ID'].iloc[0]
        ]

        # check there were few failed models in this simple setup (fancier models are expected to fail sometimes!)
        success_rate = initial_results['Exceptions'].isnull().mean()
        allowed_patterns = (
            'Transformer Detrend failed on fit',
            'Transformer bkfilter failed on fit',
            "Tensorflow not available",
            "No module named 'tensorflow'",
            "regression_type='User' but no future_regressor passed",
        )
        exception_text = initial_results['Exceptions'].dropna().astype(str)
        unexpected = [
            exc
            for exc in exception_text
            if not any(pat in exc for pat in allowed_patterns)
        ]
        unexpected_rate = len(unexpected) / len(initial_results)
        self.assertGreater(
            success_rate,
            0.7,
            f"Too many models failed ({success_rate:.2%} success rate).",
        )
        self.assertLess(
            unexpected_rate,
            0.15,
            f"Unexpected failures occurred: {unexpected[:5]}",
        )
        # check general model setup
        # self.assertEqual(validated_count, model.models_to_validate)
        self.assertGreater(
            model.validation_template.size,
            (initial_results['ValidationRound'] == 0).sum() * models_to_validate - 2,
        )
        self.assertEqual(
            set(initial_results['Model'].unique().tolist())
            - {'Ensemble', 'MLEnsemble'},
            set(model.model_list),
        )
        self.assertFalse(model.best_model.empty)
        # check the generated forecasts look right
        self.assertEqual(forecasts_df.shape[0], forecast_length)
        self.assertEqual(forecasts_df.shape[1], df.shape[1])
        self.assertFalse(forecasts_df.isna().any().any())
        self.assertTrue((forecasts_df >= 0).all().all())
        self.assertEqual(forecast_length, len(forecasts_df.index))
        self.assertTrue((expected_idx == pd.DatetimeIndex(forecasts_df.index)).all())
        # these next two could potentiall fail if any inputs have a strong trend
        self.assertTrue((forecasts_df.mean() <= (df.max()) + df.std()).all())
        self.assertTrue((forecasts_df.mean() >= (df.min()) - df.std()).all())
        # check all the checks work
        self.assertEqual(model.ensemble_check, 1)
        self.assertFalse(model.weighted)
        self.assertFalse(model.subset_flag)
        # assess 'backwards' validation
        self.assertEqual(len(model.validation_test_indexes), num_validations + 1)
        self.assertTrue(
            model.validation_test_indexes[1]
            .intersection(model.validation_train_indexes[1])
            .empty
        )
        self.assertTrue(
            model.validation_test_indexes[2]
            .intersection(model.validation_train_indexes[2])
            .empty
        )
        self.assertEqual(
            model.validation_train_indexes[1].shape[0],
            df.shape[0] - (forecast_length * 2 + 1),
        )  # +1 via drop most recent
        self.assertTrue((model.validation_test_indexes[1] == expected_val1).all())
        self.assertTrue((model.validation_test_indexes[2] == expected_val2).all())
        # assess Horizontal Ensembling
        tested_horizontal = 'horizontal' in template_dict['model_name'].lower()
        tested_mosaic = 'mosaic' in template_dict['model_name'].lower()
        print(
            f"chosen model was mosaic: {tested_mosaic} or was horizontal: {tested_horizontal}"
        )
        self.assertTrue(tested_horizontal or tested_mosaic)
        self.assertEqual(len(template_dict['series'].keys()), df.shape[1])
        if tested_horizontal:
            self.assertEqual(
                len(set(template_dict['series'].values())), template_dict['model_count']
            )
        self.assertEqual(
            len(template_dict['models'].keys()), template_dict['model_count']
        )
        # check that the create number of models were available that were requested
        one_mos = initial_results[
            initial_results["ModelParameters"].str.contains("mosaic-spl-3-10")
        ]
        res = []
        for x in json.loads(one_mos["ModelParameters"].iloc[0])["series"].values():
            for y in x.values():
                res.append(y)
        self.assertLessEqual(len(set(res)), 10)
        # check all mosaic and horizontal styles were created
        count_horz = len([x for x in ensemble if "horizontal" in x or "mosaic" in x])
        actual_unique = initial_results[initial_results["Ensemble"] == 2][
            "ModelParameters"
        ].unique()
        self.assertEqual(
            len(actual_unique),
            count_horz,
            msg=f"Expected {count_horz} horizontal/mosaic ensembles but found {len(actual_unique)}. Expected ensemble types: {[x for x in ensemble if 'horizontal' in x or 'mosaic' in x]}",
        )
        # check the mosaic details were equal
        self.assertTrue(
            len(model.initial_results.full_mae_errors)
            == len(model.initial_results.full_mae_ids)
            == len(model.initial_results.full_mae_vals)
        )
        # check at least 1 'simple' ensemble worked
        self.assertGreater(
            initial_results[initial_results["Ensemble"] == 1]['Exceptions']
            .isnull()
            .sum(),
            0,
        )
        # test that actually the best model (or nearly) was chosen
        self.assertGreater(
            validation_results['Score'].quantile(0.05),
            best_model_result['Score'].iloc[0],
        )
        # test back_forecast
        # self.assertTrue((back_forecast.index == model.df_wide_numeric.index).all(), msg="Back forecasting failed to have equivalent index to train.")
        self.assertFalse(np.any(back_forecast.isnull()))
        self.assertEqual(
            long_form.shape[0], forecasts_df.shape[0] * forecasts_df.shape[1] * 3
        )
        # results present
        self.assertGreater(model.initial_results.per_series_metrics.shape[0], 1)
        # assert that per_series results have appropriate column names
        self.assertCountEqual(
            model.initial_results.per_series_mae.columns.tolist(), df.columns.tolist()
        )

        # TEST EXPORTING A TEMPLATE THEN USING THE BEST MODEL AS A PREDICTION
        df_train = df.iloc[:-forecast_length]
        df_test = df.iloc[-forecast_length:]
        tf = tempfile.NamedTemporaryFile(
            suffix='.csv', prefix=os.path.basename("autots_test"), delete=False
        )
        time.sleep(1)
        name = tf.name
        model.export_template(name, models="best", n=20, max_per_model_class=3)
        future_regressor_train2d, future_regressor_forecast2d = fake_regressor(
            df,
            dimensions=4,
            forecast_length=forecast_length,
            date_col='datetime' if long else None,
            value_col='value' if long else None,
            id_col='series_id' if long else None,
            drop_most_recent=0,
            aggfunc=model.aggfunc,
            verbose=model.verbose,
        )

        model2 = AutoTS(
            forecast_length=forecast_length,
            frequency='infer',
            prediction_interval=0.9,
            ensemble='all',
            constraint=None,
            max_generations=generations,
            num_validations=num_validations,
            validation_method=validation_method,
            model_list="update_fit",
            transformer_list=transformer_list,
            transformer_max_depth=transformer_max_depth,
            initial_template='General+Random',
            metric_weighting=metric_weighting,
            models_to_validate=models_to_validate,
            max_per_model_class=None,
            model_interrupt=False,
            no_negatives=True,
            subset=100,
            n_jobs=n_jobs,
            drop_most_recent=0,
            verbose=2,
        )
        # TEST MODEL PREDICT WITH LOWER LEVEL MODEL TRAINED ON PREVIOUS DATA ONLY
        model2.import_best_model(tf.name, include_ensemble=False)
        model2.fit_data(
            df_train, future_regressor=future_regressor_train2d.reindex(df_train.index)
        )
        prediction = model2.predict(
            future_regressor=future_regressor_train2d.reindex(df_test.index), verbose=0
        )
        prediction.evaluate(df_test, df_train=df_train)
        smape1 = prediction.avg_metrics['smape']

        model2.fit_data(df, future_regressor=future_regressor_train2d)
        try:
            prediction2 = model2.predict(
                future_regressor=future_regressor_forecast2d, verbose=0
            )
        except Exception as e:
            raise ValueError(f"prediction failed with model {model2.best_model}") from e
        forecasts_df2 = prediction2.forecast

        # now retrain on full data
        model2.model = None
        model2.fit_data(df, future_regressor=future_regressor_train2d)
        prediction2 = model2.predict(
            future_regressor=future_regressor_forecast2d, verbose=0
        )
        # and see if it got better on past holdout
        model2.fit_data(
            df_train, future_regressor=future_regressor_train2d.reindex(df_train.index)
        )
        prediction = model2.predict(
            future_regressor=future_regressor_train2d.reindex(df_test.index), verbose=0
        )
        prediction.evaluate(df_test, df_train=df_train)
        smape2 = prediction.avg_metrics['smape']
        print("=====================================================")
        # smape2 should be better because it is trained on the very data it is supposed to predict
        print(f"fit 1 SMAPE {smape1}, then refit with history SMAPE: {smape2}")

        tf.close()
        os.unlink(tf.name)

        self.assertEqual(forecasts_df2.shape[0], forecast_length)
        self.assertEqual(forecasts_df2.shape[1], df.shape[1])
        self.assertFalse(forecasts_df2.isna().any().any())
        self.assertFalse(model.initial_results.per_series_metrics.empty)

    def test_all_default_models(self):
        print("Starting test_all_default_models")
        forecast_length = 8
        long = False
        df = load_daily(long=long).drop(
            columns=['US.Total.Covid.Tests'], errors='ignore'
        )
        # to make it faster
        df = df[df.columns[0:2]]
        n_jobs = 'auto'
        verbose = -1
        validation_method = "backwards"
        generations = 1
        num_validations = 1
        models_to_validate = 0.10  # must be a decimal percent for this test

        model_list = "default"

        transformer_list = "fast"  # ["SinTrend", "MinMaxScaler"]
        transformer_max_depth = 1
        constraint = {
            "constraint_method": "quantile",
            "constraint_regularization": 0.9,
            "upper_constraint": 0.99,
            "lower_constraint": 0.01,
            "bounds": True,
        }

        model = AutoTS(
            forecast_length=forecast_length,
            frequency='infer',
            prediction_interval=0.9,
            ensemble=["horizontal-max"],
            constraint=constraint,
            max_generations=generations,
            num_validations=num_validations,
            validation_method=validation_method,
            model_list=model_list,
            transformer_list=transformer_list,
            transformer_max_depth=transformer_max_depth,
            initial_template='Random',
            models_to_validate=models_to_validate,
            max_per_model_class=None,
            n_jobs=n_jobs,
            model_interrupt=True,
            drop_most_recent=1,
            verbose=verbose,
            random_seed=1918,
        )
        model = model.fit(
            df,
            date_col='datetime' if long else None,
            value_col='value' if long else None,
            id_col='series_id' if long else None,
        )
        prediction = model.predict(verbose=0)
        forecasts_df = prediction.forecast
        initial_results = model.results()
        validation_results = model.results("validation")

        # validated_count = (validation_results['Runs'] == (num_validations + 1)).sum()
        validated_count = (validation_results['Runs'] > 1).sum()

        # so these account for DROP MOST RECENT = 1
        expected_idx = pd.date_range(
            start=df.index[-2], periods=forecast_length + 1, freq='D'
        )[1:]
        expected_val1 = pd.date_range(
            end=df.index[-(forecast_length + 2)], periods=forecast_length, freq='D'
        )

        template_dict = json.loads(model.best_model['ModelParameters'].iloc[0])
        best_model_result = validation_results[
            validation_results['ID'] == model.best_model['ID'].iloc[0]
        ]

        check_fails = initial_results.groupby("Model")["mae"].count() > 0

        # check that all models had at least 1 success
        self.assertEqual(
            set(initial_results['Model'].unique().tolist()) - {'Ensemble'},
            set(default_model_list),
            msg="Not all models used in initial template.",
        )
        self.assertTrue(
            check_fails.all(),
            msg=f"These models failed: {check_fails[~check_fails].index.tolist()}. It is more likely a package install problem than a code problem",
        )
        # check general model setup
        self.assertGreaterEqual(validated_count, model.models_to_validate)
        lvl1 = initial_results[initial_results["Exceptions"].isnull()]
        self.assertGreater(
            model.models_to_validate,
            (lvl1[lvl1["Ensemble"] == 0]['ValidationRound'] == 0).sum()
            * models_to_validate
            - 1,
        )
        self.assertFalse(model.best_model.empty)
        # check the generated forecasts look right
        self.assertEqual(forecasts_df.shape[0], forecast_length)
        self.assertEqual(forecasts_df.shape[1], df.shape[1])
        self.assertFalse(forecasts_df.isna().any().any())
        self.assertEqual(forecast_length, len(forecasts_df.index))
        self.assertTrue((expected_idx == pd.DatetimeIndex(forecasts_df.index)).all())
        # these next two could potentiall fail if any inputs have a strong trend
        self.assertTrue((forecasts_df.mean() <= (df.max()) + df.std()).all())
        self.assertTrue((forecasts_df.mean() >= (df.min()) - df.std()).all())
        # check all the checks work
        self.assertEqual(model.ensemble_check, 1)
        self.assertFalse(model.weighted)
        self.assertFalse(model.subset_flag)
        self.assertFalse(model.used_regressor_check)
        # assess 'backwards' validation
        val_1 = model.validation_test_indexes[1]
        self.assertEqual(len(model.validation_test_indexes), num_validations + 1)
        self.assertTrue(val_1.intersection(model.validation_train_indexes[1]).empty)
        self.assertEqual(
            model.validation_train_indexes[1].shape[0],
            df.shape[0] - (forecast_length * 2 + 1),
        )  # +1 via drop most recent
        self.assertTrue((val_1 == expected_val1).all())
        # assess Horizontal Ensembling
        self.assertTrue('horizontal' in template_dict['model_name'].lower())
        self.assertEqual(len(template_dict['series'].keys()), df.shape[1])
        self.assertEqual(
            len(set(template_dict['series'].values())), template_dict['model_count']
        )
        self.assertEqual(
            len(template_dict['models'].keys()), template_dict['model_count']
        )
        # test that actually the best model (or nearly) was chosen
        self.assertGreater(
            validation_results[validation_results['Runs'] > model.num_validations][
                'Score'
            ].quantile(0.05),
            best_model_result['Score'].iloc[0],
            model.best_model.iloc[0]["Model"],
        )
        # test metrics
        self.assertTrue(initial_results['Score'].min() > 0)
        self.assertTrue(initial_results['mae'].min() >= 0)
        self.assertTrue(initial_results['smape'].min() >= 0)
        self.assertTrue(initial_results['rmse'].min() >= 0)
        self.assertTrue(initial_results['contour'].min() >= 0)
        self.assertTrue(initial_results['containment'].min() >= 0)
        self.assertTrue(initial_results['TotalRuntimeSeconds'].min() >= 0)
        self.assertTrue(initial_results['spl'].min() >= 0)
        self.assertTrue(initial_results['contour'].min() <= 1)
        self.assertTrue(initial_results['containment'].min() <= 1)

        # Test that generate_score is actually picking the lowest value on a single metric
        # minimizing metrics only
        # no weights present on metrics
        for target_metric in [
            'smape',
            'mae',
            'rmse',
            'made',
            'spl',
            'mage',
            'mle',
            'imle',
            'dwae',
            'mqae',
            'uwmse',
            'wasserstein',
            'dwd',
        ]:
            with self.subTest(i=target_metric):
                new_weighting = {
                    str(target_metric) + '_weighting': 1,
                }
                temp_cols = [
                    'ID',
                    'Model',
                    'ModelParameters',
                    'TransformationParameters',
                    'Ensemble',
                    target_metric,
                ]
                new_mod = model._return_best_model(
                    metric_weighting=new_weighting, template_cols=temp_cols
                )
                new_mod_non = new_mod[1]
                new_mod = new_mod[0]
                if new_mod['Ensemble'].iloc[0] == 2:
                    min_pos = validation_results[validation_results['Ensemble'] == 2][
                        target_metric
                    ].min()
                    min_pos_non = validation_results[
                        (validation_results['Ensemble'] < 2)
                        & (validation_results['Runs'] > model.num_validations)
                    ][target_metric].min()
                    chos_pos = new_mod[target_metric].iloc[0]
                    # print(min_pos)
                    # print(chos_pos)
                    self.assertTrue(np.allclose(chos_pos, min_pos))
                    self.assertTrue(
                        np.allclose(new_mod_non[target_metric].iloc[0], min_pos_non)
                    )
                    # print(json.loads(new_mod['ModelParameters'].iloc[0])['model_name'])
                    # print(json.loads(new_mod['ModelParameters'].iloc[0])['model_metric'])

    def test_load_datasets(self):
        df = load_artificial(long=True)
        df = load_monthly(long=True)
        df = load_sine(long=False)
        df = load_daily(long=False)
        df = load_daily(long=True)  # noqa

    def test_new_params(self):
        params = AutoTS.get_new_params()
        self.assertIsInstance(params, dict)
        model = AutoTS(**params)  # noqa

        params = AutoTS.get_new_params(method='fast')
        self.assertIsInstance(params, dict)
        model = AutoTS(**params)  # noqa

    def test_univariate1step(self):
        print("Starting test_univariate1step")
        df = load_artificial(long=False)
        df.iloc[:, :1]
        forecast_length = 1
        n_jobs = 1
        verbose = -1
        validation_method = "backwards"
        generations = 1
        model_list = [
            'ConstantNaive',
            'LastValueNaive',
            'AverageValueNaive',
            'SeasonalNaive',
        ]

        model = AutoTS(
            forecast_length=forecast_length,
            frequency='infer',
            max_generations=generations,
            validation_method=validation_method,
            model_list=model_list,
            n_jobs=n_jobs,
            verbose=verbose,
        )
        model = model.fit(
            df,
        )
        prediction = model.predict(verbose=0)
        forecasts_df = prediction.forecast
        initial_results = model.results()

        expected_idx = pd.date_range(
            start=df.index[-1], periods=forecast_length + 1, freq='D'
        )[1:]
        check_fails = initial_results.groupby("Model")["mae"].count() > 0
        self.assertTrue(
            check_fails.all(),
            msg=f"These models failed: {check_fails[~check_fails].index.tolist()}. It is more likely a package install problem than a code problem",
        )
        # check the generated forecasts look right
        self.assertEqual(forecasts_df.shape[0], forecast_length)
        self.assertEqual(forecasts_df.shape[1], df.shape[1])
        self.assertFalse(forecasts_df.isna().any().any())
        self.assertEqual(forecast_length, len(forecasts_df.index))
        self.assertTrue((expected_idx == pd.DatetimeIndex(forecasts_df.index)).all())

    def test_subset_expansion(self):
        # probably has room for testing some more things as well
        long = True
        df = load_daily(long=long)
        forecast_length = 28
        model = AutoTS(
            forecast_length=forecast_length,
            frequency='infer',
            max_generations=10,
            validation_method="seasonal",
            model_list="superfast",
            ensemble=[
                "horizontal-max",
                "mosaic-weighted-0-10",
                "mosaic-mae-crosshair-0-20",
            ],
            n_jobs=2,
            verbose=-1,
            subset=4,
            remove_leading_zeroes=True,
        )
        model = model.fit(
            df,
            date_col="datetime" if long else None,
            value_col="value" if long else None,
            id_col="series_id" if long else None,
        )
        model.expand_horizontal()
        orig_param = json.loads(model.best_model_original.iloc[0]['ModelParameters'])
        new_param = json.loads(model.best_model.iloc[0]['ModelParameters'])
        diff_mods = [
            x
            for x in orig_param['models'].keys()
            if x not in new_param['models'].keys()
        ]
        if diff_mods:
            details = orig_param['models'][diff_mods[0]]
        else:
            details = ""
        self.assertCountEqual(
            orig_param['models'].keys(),
            new_param['models'].keys(),
            msg=f"model expansion failed to use the same models {details}",
        )
        num_series = len(df['series_id'].unique().tolist()) if long else df.shape[1]
        self.assertEqual(
            len(
                json.loads(model.best_model.iloc[0]['ModelParameters'])['series'].keys()
            ),
            num_series,
            msg="model expansion failed to expand to all df columns",
        )
        prediction = model.predict(verbose=0)
        forecasts_df = prediction.forecast
        initial_results = model.results()

        check_fails = initial_results.groupby("Model")["mae"].count() > 0
        self.assertTrue(
            check_fails.all(),
            msg=f"These models failed: {check_fails[~check_fails].index.tolist()}. It is more likely a package install problem than a code problem",
        )
        # check the generated forecasts look right
        self.assertEqual(forecasts_df.shape[0], forecast_length)
        self.assertEqual(forecasts_df.shape[1], num_series)
        self.assertFalse(forecasts_df.isna().any().any())
        self.assertEqual(forecast_length, len(forecasts_df.index))

    def test_mixedlengthval(self):
        print("Starting test_mixedlengthval")
        df = load_artificial(long=False)
        forecast_length = 2
        n_jobs = 1
        verbose = -1
        validation_method = "mixed_length"
        generations = 1
        model_list = [
            'ConstantNaive',
            'LastValueNaive',
            'AverageValueNaive',
            'SeasonalNaive',
        ]

        def custom_metric(A, F, df_train=None, prediction_interval=None):
            submission = F
            objective = A
            abs_err = np.nansum(np.abs(submission - objective))
            err = np.nansum((submission - objective))
            score = abs_err + abs(err)
            epsilon = 1
            big_sum = (
                np.nan_to_num(objective, nan=0.0, posinf=0.0, neginf=0.0).sum().sum()
                + epsilon
            )
            score /= big_sum
            return score

        model = AutoTS(
            forecast_length=forecast_length,
            frequency='infer',
            max_generations=generations,
            validation_method=validation_method,
            model_list=model_list,
            n_jobs=n_jobs,
            verbose=verbose,
            ensemble=None,
            custom_metric=custom_metric,
        )
        model = model.fit(
            df,
        )
        prediction = model.predict(verbose=0)
        forecasts_df = prediction.forecast
        initial_results = model.results()

        expected_idx = pd.date_range(
            start=df.index[-1], periods=forecast_length + 1, freq='D'
        )[1:]
        check_fails = initial_results.groupby("Model")["mae"].count() > 0
        self.assertTrue(
            check_fails.all(),
            msg=f"These models failed: {check_fails[~check_fails].index.tolist()}. It is more likely a package install problem than a code problem",
        )
        # check the generated forecasts look right
        self.assertEqual(forecasts_df.shape[0], forecast_length)
        self.assertEqual(forecasts_df.shape[1], df.shape[1])
        self.assertFalse(forecasts_df.isna().any().any())
        self.assertEqual(forecast_length, len(forecasts_df.index))
        self.assertTrue((expected_idx == pd.DatetimeIndex(forecasts_df.index)).all())
        # test custom metric is non-zero
        self.assertNotAlmostEqual(initial_results["custom"].mean(), 0)

    def test_all_models_load(self):
        print("Starting test_all_models_load")
        # make sure it can at least load a template of all models
        forecast_length = 8
        n_jobs = 'auto'
        verbose = 4
        generations = 0

        model_list = "all"
        transformer_list = "all"  # ["SinTrend", "MinMaxScaler"]
        transformer_max_depth = 10

        model = AutoTS(
            forecast_length=forecast_length,
            frequency='infer',
            prediction_interval=0.9,
            ensemble=["horizontal-max"],
            constraint=None,
            max_generations=generations,
            model_list=model_list,
            transformer_list=transformer_list,
            transformer_max_depth=transformer_max_depth,
            initial_template='Random',
            max_per_model_class=None,
            n_jobs=n_jobs,
            model_interrupt=True,
            drop_most_recent=1,
            verbose=verbose,
        )
        self.assertFalse(model.initial_template.empty)

    def test_benchmark(self):
        print("Starting test_benchmark")
        bench = Benchmark()
        bench.run(times=1, verbose=-1)
        self.assertGreater(bench.total_runtime, 0)
        print(f"Benchmark total_runtime: {bench.total_runtime}")
        print(bench.results)
        time.sleep(5)

    def test_template(self):
        for index, row in general_template.iterrows():
            model = row["Model"]
            with self.subTest(i=model):
                mod = json.loads(row['ModelParameters'])
                trans = json.loads(row['TransformationParameters'])
                ensemble = row["Ensemble"]
                self.assertIsInstance(mod, dict)
                self.assertIsInstance(trans, dict)
                self.assertIsNotNone(ensemble)

    def test_custom_validations(self):
        long = False
        df = load_daily(long=long)
        forecast_length = 28
        model = AutoTS(
            forecast_length=forecast_length,
            frequency='D',
            max_generations=10,
            validation_method="custom",
            model_list="superfast",
            ensemble=None,
            n_jobs=1,
            verbose=2,
            subset=4,
            remove_leading_zeroes=True,
            generation_timeout=1,
            horizontal_ensemble_validation=True,
        )
        custom_idx = [
            pd.date_range(df.index[0], df.index[-100]),
            pd.date_range(df.index[0], df.index[-(100 + forecast_length)]),
        ]
        model = model.fit(df, validation_indexes=custom_idx)
        self.assertEqual(model.ensemble_check, 0)

        # test all same on univariate input, non-horizontal, with regressor, and different frequency, with forecast_length = 1 !

        # the big ones are:
        # 1. that validations are sampled correctly
        # 2. that accuracy metrics are performed and aggregated correctly

        # test template import and export
        # test result saving and import
        # test seasonal validation
        # test score generation + metric_weighting
        # test very short training data and/or lots of NaNs in data
        # test on all models that for each model, failure rate is < 100%


transforms = [
    'MinMaxScaler',
    'PowerTransformer',
    'QuantileTransformer',
    'MaxAbsScaler',
    'StandardScaler',
    'RobustScaler',
    'PCA',
    'FastICA',
    "DatepartRegression",
    "EWMAFilter",
    'STLFilter',
    'HPFilter',
    'Detrend',
    'Slice',
    'ScipyFilter',
    'Round',
    'ClipOutliers',
    'IntermittentOccurrence',
    'CenterLastValue',
    'Discretize',
    'SeasonalDifference',
    'RollingMeanTransformer',
    'bkfilter',
    'cffilter',
    'Log',
    'DifferencedTransformer',
    'PctChangeTransformer',
    'PositiveShift',
    'SineTrend',
    'convolution_filter',
    'CumSumTransformer',
    'AlignLastValue',  # new 0.4.3
    'AnomalyRemoval',
    "HolidayTransformer",  # new 0.5.0
    'LocalLinearTrend',  # new 0.5.1
    "KalmanSmoothing",  # new 0.5.1
    "RegressionFilter",  # new 0.5.7
    "LevelShiftTransformer",  # new 0.6.0
    "CenterSplit",  # new 0.6.1
    "FFTFilter",
    "ReplaceConstant",
    "AlignLastDiff",  # new 0.6.2
    "FFTDecomposition",  # new in 0.6.2
    "HistoricValues",  # new in 0.6.7
    "BKBandpassFilter",  # new in 0.6.8
    "Constraint",  # new in 0.6.15
    "DiffSmoother",  # new in 0.6.15
    "FIRFilter",  # new in 0.6.16
    "ShiftFirstValue",  # new in 0.6.16
    "ThetaTransformer",  # new in 0.6.16
    "ChangepointDetrend",  # new in 0.6.16
    "MeanPercentSplitter",  # new in 0.6.16
    "UpscaleDownscaleTransformer",  # new in 0.6.18
    # "ReconciliationTransformer",  # new in 0.6.22
    # "CointegrationTransformer",  # new in 0.6.22
]


class ModelTest(unittest.TestCase):
    def test_models_get_params(self):
        """See if new random params can be generated without error."""
        default_methods = [
            'deep',
            'fast',
            'random',
            'default',
            'superfast',
            'regressor',
            'event_risk',
        ]
        for method in default_methods:
            for model_str in all_models:
                ModelMonster(model_str).get_new_params(method=method)

    def test_models(self):
        """Test if models are the same as saved comparisons."""
        print("Starting test_models")
        n_jobs = 1
        random_seed = 300
        df = load_daily(long=False).iloc[:, 0:5]
        df = df[
            df.index < "2022-10-04"
        ]  # update dataset and have not yet updated stored model results
        df = df[
            df.index > "2017-10-04"
        ]  # update dataset and have not yet updated stored model results
        models = [
            'SectionalMotif',
            'MultivariateMotif',
            'AverageValueNaive',
            'NVAR',
            "LastValueNaive",
            'Theta',
            'FBProphet',
            'SeasonalNaive',
            'GLM',
            'ETS',
            "ConstantNaive",
            'WindowRegression',
            'DatepartRegression',
            'MultivariateRegression',
            'Cassandra',
            'MetricMotif',
            'SeasonalityMotif',
            'KalmanStateSpace',
            'ARDL',
            'UnivariateMotif',
            'VAR',
            'MAR',
            'TMF',
            'RRVAR',
            'VECM',
            'BallTreeMultivariateMotif',
            'FFT',
            "DMD",  # 0.6.12
            "BasicLinearModel",
            "TVVAR",  # 0.6.16
            # "BallTreeRegressionMotif",  # 0.6.17
            # "PreprocessingExperts",  # 0.6.18
        ]
        # models that for whatever reason arne't consistent across test sessions
        run_only_no_score = ['FBProphet', 'RRVAR', "TMF"]

        timings = {}
        forecasts = {}
        upper_forecasts = {}
        lower_forecasts = {}
        # load the comparison source
        with open("./tests/model_forecasts.json", "r") as file:
            loaded = json.load(file)
            for x in models:
                forecasts[x] = pd.DataFrame.from_dict(
                    loaded['forecasts'][x], orient="columns"
                )
                forecasts[x]['index'] = pd.to_datetime(forecasts[x]['index'])
                forecasts[x] = forecasts[x].set_index("index")
                upper_forecasts[x] = pd.DataFrame.from_dict(
                    loaded['upper_forecasts'][x], orient="columns"
                )
                upper_forecasts[x]['index'] = pd.to_datetime(
                    upper_forecasts[x]['index']
                )
                upper_forecasts[x] = upper_forecasts[x].set_index("index")
                lower_forecasts[x] = pd.DataFrame.from_dict(
                    loaded['lower_forecasts'][x], orient="columns"
                )
                lower_forecasts[x]['index'] = pd.to_datetime(
                    lower_forecasts[x]['index']
                )
                lower_forecasts[x] = lower_forecasts[x].set_index("index")
            timings = loaded['timing']

        timings2 = {}
        forecasts2 = {}
        upper_forecasts2 = {}
        lower_forecasts2 = {}
        # following are not consistent with seed:
        # "MotifSimulation"

        for x in models:
            print(x)
            try:
                start_time = timeit.default_timer()
                df_forecast = model_forecast(
                    model_name=x,
                    model_param_dict={},  # 'return_result_windows': True
                    model_transform_dict={
                        "fillna": "ffill",
                        "transformations": {"0": "StandardScaler"},
                        "transformation_params": {"0": {}},
                    },
                    df_train=df,
                    forecast_length=5,
                    frequency="D",
                    prediction_interval=0.9,
                    random_seed=random_seed,
                    verbose=0,
                    # bug in sklearn 1.1.2 for n_jobs for RandomForest
                    n_jobs=n_jobs if x != "WindowRegression" else 2,
                    return_model=True,
                )
                forecasts2[x] = df_forecast.forecast.round(2)
                upper_forecasts2[x] = df_forecast.upper_forecast.round(2)
                lower_forecasts2[x] = df_forecast.lower_forecast.round(2)
                timings2[x] = timeit.default_timer() - start_time
            except Exception as e:
                raise ValueError(f"model {x} failed with {repr(e)}")

        print(sum(timings.values()))

        pass_probabilistic = [
            'FBProphet'
        ]  # not yet reproducible in upper/lower with seed
        for x in models:
            if x not in run_only_no_score:
                with self.subTest(i=x):
                    res = (forecasts2[x].round(2) == forecasts[x].round(2)).all().all()
                    if x not in pass_probabilistic:
                        res_u = (
                            (
                                upper_forecasts2[x].round(2)
                                == upper_forecasts[x].round(2)
                            )
                            .all()
                            .all()
                        )
                        res_l = (
                            (
                                lower_forecasts2[x].round(2)
                                == lower_forecasts[x].round(2)
                            )
                            .all()
                            .all()
                        )
                    else:
                        res_u = True
                        res_l = True
                    self.assertTrue(
                        res,
                        f"Model '{x}' forecasts diverged from sample forecasts.",  # note that the same code in different environments may give slightly different results, WindowRegression, MultivariateRegression, BallTreeMultivariateMotif, NVAR known to have this issue
                    )
                    self.assertTrue(
                        res_u,
                        f"Model '{x}' upper forecasts diverged from sample forecasts.",
                    )
                    self.assertTrue(
                        res_l,
                        f"Model '{x}' lower forecasts diverged from sample forecasts.",
                    )
                    print(
                        f"{res & res_u & res_l} model '{x}' ran successfully in {round(timings2[x], 4)} (bench: {round(timings[x], 4)})"
                    )

        """
        for x in models:
            forecasts[x].index = forecasts[x].index.strftime("%Y-%m-%d")
            forecasts[x] = forecasts[x].reset_index(drop=False).to_dict(orient="list")
            upper_forecasts[x].index = upper_forecasts[x].index.strftime("%Y-%m-%d")
            upper_forecasts[x] = upper_forecasts[x].reset_index(drop=False).to_dict(orient="list")
            lower_forecasts[x].index = lower_forecasts[x].index.strftime("%Y-%m-%d")
            lower_forecasts[x] = lower_forecasts[x].reset_index(drop=False).to_dict(orient="list")

        with open("./tests/model_forecasts.json", "w") as file:
            json.dump(
                {
                    'forecasts': forecasts,
                    "upper_forecasts": upper_forecasts,
                    "lower_forecasts": lower_forecasts,
                    "timing": timings,
                }, file
            )
        """

    def test_transforms(self):
        """Test if transformers meet saved comparison outputs."""
        print("Starting test_transforms")
        n_jobs = 1
        random_seed = 300
        df = load_monthly(long=False)[['CSUSHPISA', 'EMVOVERALLEMV', 'EXCAUS']]

        timings = {}
        forecasts = {}
        upper_forecasts = {}
        lower_forecasts = {}
        # load the comparison source
        with open("./tests/transform_forecasts.json", "r") as file:
            loaded = json.load(file)
            for x in transforms:
                forecasts[x] = pd.DataFrame.from_dict(
                    loaded['forecasts'][x], orient="columns"
                )
                forecasts[x]['index'] = pd.to_datetime(forecasts[x]['index'])
                forecasts[x] = forecasts[x].set_index("index")
                upper_forecasts[x] = pd.DataFrame.from_dict(
                    loaded['upper_forecasts'][x], orient="columns"
                )
                upper_forecasts[x]['index'] = pd.to_datetime(
                    upper_forecasts[x]['index']
                )
                upper_forecasts[x] = upper_forecasts[x].set_index("index")
                lower_forecasts[x] = pd.DataFrame.from_dict(
                    loaded['lower_forecasts'][x], orient="columns"
                )
                lower_forecasts[x]['index'] = pd.to_datetime(
                    lower_forecasts[x]['index']
                )
                lower_forecasts[x] = lower_forecasts[x].set_index("index")
            timings = loaded['timing']

        timings2 = {}
        forecasts2 = {}
        upper_forecasts2 = {}
        lower_forecasts2 = {}

        for x in transforms:
            print(x)
            forecast_length = 5
            if x in ['QuantileTransformer']:
                param = {"n_quantiles": 100}
            elif x in ["UpscaleDownscaleTransformer"]:
                param = {
                    'mode': 'upscale',
                    'factor': 3,
                    'down_method': 'decimate',
                    'fill_method': 'pchip',
                }
            else:
                param = {}
            start_time = timeit.default_timer()
            model = ModelPrediction(
                forecast_length=forecast_length,
                transformation_dict={
                    "fillna": "ffill",
                    "transformations": {"0": x},
                    "transformation_params": {"0": param},
                },
                model_str="LastValueNaive",
                parameter_dict={},
                frequency="infer",
                prediction_interval=0.9,
                random_seed=random_seed,
                verbose=-1,
                fail_on_forecast_nan=True,
                n_jobs=n_jobs,
                return_model=True,
            )
            model = model.fit(df)
            df_forecast = model.predict(forecast_length=forecast_length)
            forecasts2[x] = df_forecast.forecast.round(2)
            upper_forecasts2[x] = df_forecast.upper_forecast.round(2)
            lower_forecasts2[x] = df_forecast.lower_forecast.round(2)
            timings2[x] = timeit.default_timer() - start_time

        print(sum(timings2.values()))

        pass_probabilistic = ['FastICA']  # not reproducible in upper/lower with seed
        for x in transforms:
            with self.subTest(i=x):
                res = (forecasts2[x].round(2) == forecasts[x].round(2)).all().all()
                if x not in pass_probabilistic:
                    res_u = (
                        (upper_forecasts2[x].round(2) == upper_forecasts[x].round(2))
                        .all()
                        .all()
                    )
                    res_l = (
                        (lower_forecasts2[x].round(2) == lower_forecasts[x].round(2))
                        .all()
                        .all()
                    )
                else:
                    res_u = True
                    res_l = True
                self.assertTrue(
                    res, f"Model '{x}' forecasts diverged from sample forecasts."
                )
                self.assertTrue(
                    res_u,
                    f"Model '{x}' upper forecasts diverged from sample forecasts.",
                )
                self.assertTrue(
                    res_l,
                    f"Model '{x}' lower forecasts diverged from sample forecasts.",
                )
                print(
                    f"{res & res_u & res_l} model '{x}' ran successfully in {round(timings2[x], 4)} (bench: {round(timings[x], 4)})"
                )

        """
        for x in transforms:
            forecasts[x].index = forecasts[x].index.strftime("%Y-%m-%d")
            forecasts[x] = forecasts[x].reset_index(drop=False).to_dict(orient="list")
            upper_forecasts[x].index = upper_forecasts[x].index.strftime("%Y-%m-%d")
            upper_forecasts[x] = upper_forecasts[x].reset_index(drop=False).to_dict(orient="list")
            lower_forecasts[x].index = lower_forecasts[x].index.strftime("%Y-%m-%d")
            lower_forecasts[x] = lower_forecasts[x].reset_index(drop=False).to_dict(orient="list")

        with open("./tests/transform_forecasts.json", "w") as file:
            json.dump(
                {
                    'forecasts': forecasts,
                    "upper_forecasts": upper_forecasts,
                    "lower_forecasts": lower_forecasts,
                    "timing": timings,
                }, file
            )
        """

    def test_transforms_get_params(self):
        """See if new random params can be generated without error."""
        from autots.datasets import load_daily

        # Create a sample DataFrame for transformers that need one
        df = load_daily(long=False).iloc[:50, :3]  # Small sample for testing

        default_methods = ['deep', 'fast', 'random', 'default', 'superfast']
        gtrans = GeneralTransformer()

        # External transformers (from sklearn) don't have get_new_params
        external_transformers = [
            "MinMaxScaler",
            "PowerTransformer",
            "QuantileTransformer",
            "MaxAbsScaler",
            "StandardScaler",
            "RobustScaler",
        ]

        for method in default_methods:
            for transform_str in transforms:
                with self.subTest(i=transform_str):
                    # Skip external transformers as they don't have get_new_params
                    if transform_str in external_transformers:
                        continue

                    trans = gtrans.retrieve_transformer(transform_str, df=df)
                    # Only test get_new_params if the transformer has this method
                    if hasattr(trans, 'get_new_params'):
                        params = trans.get_new_params(method=method)
                        self.assertIsInstance(params, dict)
                        trans = gtrans.retrieve_transformer(
                            transform_str, param=params, df=df
                        )

    def test_sklearn(self):
        from autots import load_daily
        from autots import create_regressor
        from autots.models.sklearn import (
            MultivariateRegression,
            DatepartRegression,
            WindowRegression,
        )

        df = load_daily(long=False).bfill().ffill()
        forecast_length = 8
        df_train = df.iloc[:-forecast_length]
        df_test = df.iloc[-forecast_length:]
        future_regressor_train, future_regressor_forecast = create_regressor(
            df_train,
            forecast_length=forecast_length,
            frequency="infer",
            drop_most_recent=0,
            scale=True,
            summarize="auto",
            backfill="bfill",
            fill_na="spline",
            holiday_countries={"US": None},  # requires holidays package
            encode_holiday_type=True,
        )

        random_seed = 300
        frequency = 'D'
        prediction_interval = 0.9
        verbose = -1
        n_jobs = 2

        params = MultivariateRegression().get_new_params()
        params = {
            'regression_model': {
                'model': 'LightGBM',
                'model_params': {
                    'objective': 'regression',
                    'learning_rate': 0.1,
                    'num_leaves': 31,
                    'max_depth': 10,
                    'boosting_type': 'goss',
                    'n_estimators': 250,
                    'linear_tree': False,
                },
            },
            'mean_rolling_periods': 90,
            'macd_periods': 12,
            'std_rolling_periods': 7,
            'max_rolling_periods': None,
            'min_rolling_periods': None,
            'quantile90_rolling_periods': 7,
            'quantile10_rolling_periods': 10,
            'ewm_alpha': 0.8,
            'ewm_var_alpha': None,
            'additional_lag_periods': None,
            'abs_energy': False,
            'rolling_autocorr_periods': None,
            'datepart_method': 'expanded',
            'polynomial_degree': None,
            'regression_type': None,
            'window': 3,
            'holiday': True,
            'probabilistic': False,
            'cointegration': None,
            'cointegration_lag': 1,
        }
        model = MultivariateRegression(
            forecast_length=forecast_length,
            frequency=frequency,
            prediction_interval=prediction_interval,
            random_seed=random_seed,
            verbose=verbose,
            n_jobs=n_jobs,
            **params,
        )
        model.fit(df_train)
        first_forecast = model.predict(future_regressor=future_regressor_forecast)
        self.assertListEqual(
            first_forecast.forecast.index.tolist(), df_test.index.tolist()
        )
        model.fit_data(df)
        updated_forecast = model.predict()
        self.assertEqual(updated_forecast.forecast.shape[0], forecast_length)
        self.assertTrue(updated_forecast.forecast.index[0] > df.index[-1])

        params = WindowRegression().get_new_params()
        params = {}
        model = WindowRegression(
            forecast_length=forecast_length,
            frequency=frequency,
            prediction_interval=prediction_interval,
            random_seed=random_seed,
            verbose=verbose,
            n_jobs=n_jobs,
            **params,
        )
        model.fit(df_train)
        first_forecast = model.predict(future_regressor=future_regressor_forecast)
        # first_forecast.plot_grid(df)
        self.assertListEqual(
            first_forecast.forecast.index.tolist(), df_test.index.tolist()
        )
        model.fit_data(df)
        updated_forecast = model.predict()
        # updated_forecast.plot_grid(df)
        self.assertEqual(updated_forecast.forecast.shape[0], forecast_length)
        self.assertTrue(updated_forecast.forecast.index[0] > df.index[-1])

        params = {
            'regression_model': {
                'model': 'ExtraTrees',
                'model_params': {
                    'n_estimators': 500,
                    'min_samples_leaf': 1,
                    'max_depth': 20,
                },
            },
            'datepart_method': 'simple_binarized',
            'polynomial_degree': None,
            'regression_type': None,
        }
        model = DatepartRegression(
            forecast_length=forecast_length,
            frequency=frequency,
            prediction_interval=prediction_interval,
            random_seed=random_seed,
            verbose=verbose,
            n_jobs=n_jobs,
            **params,
        )
        model.fit(df_train)
        first_forecast = model.predict(future_regressor=future_regressor_forecast)
        self.assertListEqual(
            first_forecast.forecast.index.tolist(), df_test.index.tolist()
        )
        model.fit_data(df)
        updated_forecast = model.predict()
        self.assertEqual(updated_forecast.forecast.shape[0], forecast_length)
        self.assertTrue(updated_forecast.forecast.index[0] > df.index[-1])

    def test_interrupt_helpers(self):
        reset_interrupt_tracking()
        first_press = _register_interrupt_press(1.5)
        self.assertFalse(first_press)
        second_press = _register_interrupt_press(1.5)
        self.assertTrue(second_press)
        reset_interrupt_tracking()
        mode, window = _normalize_interrupt_mode(
            {"mode": "skip", "double_press_window": 1.2}
        )
        self.assertEqual(mode, "skip")
        self.assertAlmostEqual(window, 1.2, places=4)
        mode_stop, _ = _normalize_interrupt_mode("stop")
        self.assertEqual(mode_stop, "run")

    def test_run_interrupt_flags(self):
        model = AutoTS(
            forecast_length=1,
            max_generations=0,
            num_validations=0,
            model_interrupt=True,
            verbose=0,
            model_list='superfast',
        )
        model.used_frequency = 'D'
        template = pd.DataFrame(columns=model.template_cols)
        idx = pd.date_range("2020-01-01", periods=2, freq='D')
        df_train = pd.DataFrame({'a': [1.0]}, index=idx[:1])
        df_test = pd.DataFrame({'a': [2.0]}, index=idx[1:])
        fake_result = TemplateEvalObject()
        fake_result.interrupted = True
        fake_result.interrupt_details = {"level": "run", "timestamp": "now"}
        with patch('autots.evaluator.auto_ts.TemplateWizard', return_value=fake_result):
            model._run_template(
                template,
                df_train,
                df_test,
                future_regressor_train=None,
                future_regressor_test=None,
                current_weights={'a': 1},
                validation_round=0,
                max_generations="0",
                model_count=0,
                result_file=None,
                return_template=True,
            )
        self.assertTrue(model._interrupt_run)
        self.assertTrue(model.run_was_interrupted)
        self.assertEqual(model.run_interrupt_details.get("level"), "run")

    def test_generation_interrupt_flags(self):
        model = AutoTS(
            forecast_length=1,
            max_generations=0,
            num_validations=0,
            model_interrupt="end_generation",
            verbose=0,
            model_list='superfast',
        )
        model.used_frequency = 'D'
        template = pd.DataFrame(columns=model.template_cols)
        idx = pd.date_range("2020-01-01", periods=2, freq='D')
        df_train = pd.DataFrame({'a': [1.0]}, index=idx[:1])
        df_test = pd.DataFrame({'a': [2.0]}, index=idx[1:])
        fake_result = TemplateEvalObject()
        fake_result.interrupted = True
        fake_result.interrupt_details = {"level": "generation", "timestamp": "now"}
        with patch('autots.evaluator.auto_ts.TemplateWizard', return_value=fake_result):
            model._run_template(
                template,
                df_train,
                df_test,
                future_regressor_train=None,
                future_regressor_test=None,
                current_weights={'a': 1},
                validation_round=0,
                max_generations="0",
                model_count=0,
                result_file=None,
                return_template=True,
            )
        self.assertFalse(model._interrupt_run)
        self.assertFalse(model.run_was_interrupted)
        self.assertEqual(model.run_interrupt_details.get("level"), "generation")

    def test_corecount(self):
        auto_count = cpu_count()
        half = int(auto_count * 0.5) if int(auto_count * 0.5) > 1 else 1
        self.assertEqual(half, set_n_jobs(0.5))
        self.assertGreater(auto_count, 0)
        self.assertGreater(set_n_jobs(-4), 0)
        self.assertEqual(set_n_jobs(8.0), 8)
        self.assertIsInstance(set_n_jobs("auto"), int)
        self.assertEqual(set_n_jobs("auto"), set_n_jobs(-1))



# -*- coding: utf-8 -*-
"""Test anomalies.
Created on Mon Jul 18 16:27:48 2022

@author: Colin
"""
import unittest
import numpy as np
import pandas as pd
from autots.tools.anomaly_utils import available_methods, fast_methods
from autots.evaluator.anomaly_detector import AnomalyDetector, HolidayDetector
from autots.datasets import load_live_daily
from autots.tools.transform import expanding_transformers


def dict_loop(params):
    if 'transform_dict' in params.keys():
        x = params.get('transform_dict', {})
        if isinstance(x, dict):
            x = x.get('transformations', {})
            return x
    elif 'anomaly_detector_params' in params.keys():
        x = params.get('anomaly_detector_params', {})
        if isinstance(x, dict):
            x = x.get('transform_dict', {})
            if isinstance(x, dict):
                x = x.get('transformations', {})
                return x
    return {}


class TestAnomalies(unittest.TestCase):
    @classmethod
    def setUp(self):
        wiki_pages = [
            "Standard_deviation",  # anti-holiday
            "Christmas",
            "Thanksgiving",  # specific holiday
            "all",
        ]
        self.df = (
            load_live_daily(
                long=False,
                fred_series=None,
                tickers=None,
                trends_list=None,
                earthquake_min_magnitude=None,
                weather_stations=None,
                london_air_stations=None,
                gov_domain_list=None,
                weather_event_types=None,
                wikipedia_pages=wiki_pages,
                caiso_query=None,
                sleep_seconds=10,
            )
            .fillna(0)
            .replace(np.inf, 0)
        )

    def test_anomaly_holiday_detectors(self):
        """Combining these to minimize live data download."""
        print("Starting test_anomaly_holiday_detectors")
        tried = []
        # Check if PyTorch is available for VAEOutlier
        try:
            import torch

            torch_available = True
        except ImportError:
            torch_available = False

        # Filter methods based on torch availability
        test_methods = [
            m for m in available_methods if m != 'VAEOutlier' or torch_available
        ]

        while not all(x in tried for x in test_methods):
            params = AnomalyDetector.get_new_params(method="deep")
            # remove 'Slice' as it messes up assertions
            while any(
                item in dict_loop(params).values() for item in expanding_transformers
            ):
                params = AnomalyDetector.get_new_params(method="deep")
            with self.subTest(i=params['method']):
                print(
                    f"Starting subtest test_anomaly_holiday_detectors method={params['method']}"
                )
                tried.append(params['method'])
                mod = AnomalyDetector(output='multivariate', **params)
                num_cols = 2
                mod.detect(
                    self.df[np.random.choice(self.df.columns, num_cols, replace=False)]
                )
                # detected = mod.anomalies
                # print(params)
                # mod.plot()
                self.assertEqual(
                    mod.anomalies.shape,
                    (self.df.shape[0], num_cols),
                    msg=f"from params {params}",
                )

                mod = AnomalyDetector(output='univariate', **params)
                mod.detect(
                    self.df[np.random.choice(self.df.columns, num_cols, replace=False)]
                )
                self.assertEqual(mod.anomalies.shape, (self.df.shape[0], 1))
        # mod.plot()

        from prophet import Prophet

        tried = []
        forecast_length = 28
        holidays_detected = 0
        full_dates = self.df.index.union(
            pd.date_range(self.df.index.max(), freq="D", periods=forecast_length)
        )

        while not all(x in tried for x in fast_methods):
            params = HolidayDetector.get_new_params(method="fast")
            with self.subTest(i=params["anomaly_detector_params"]['method']):
                print(
                    f"Starting subtest test_anomaly_holiday_detectors method={params['anomaly_detector_params']['method']}"
                )
                while any(
                    item in dict_loop(params).values()
                    for item in expanding_transformers
                ):
                    params = HolidayDetector.get_new_params(method="fast")
                tried.append(params['anomaly_detector_params']['method'])
                mod = HolidayDetector(**params)
                mod.detect(self.df.copy())
                prophet_holidays = mod.dates_to_holidays(full_dates, style="prophet")

                for series in self.df.columns:
                    # series = "wiki_George_Washington"
                    holiday_subset = prophet_holidays[
                        prophet_holidays['series'] == series
                    ]
                    if holiday_subset.shape[0] >= 1:
                        holidays_detected = 1
                    m = Prophet(holidays=holiday_subset)
                    # m = Prophet()
                    m.fit(pd.DataFrame({'ds': self.df.index, 'y': self.df[series]}))
                    future = m.make_future_dataframe(forecast_length)
                    fcst = m.predict(future).set_index('ds')  # noqa
                    # m.plot_components(fcst)
        # mod.plot()
        temp = mod.dates_to_holidays(full_dates, style="flag")
        temp = mod.dates_to_holidays(full_dates, style="series_flag")
        temp = mod.dates_to_holidays(full_dates, style="impact")
        temp = mod.dates_to_holidays(full_dates, style="long")  # noqa
        # this is a weak test, but will capture some functionality
        self.assertEqual(holidays_detected, 1, "no methods detected holidays")


class TestVAEAnomalies(unittest.TestCase):
    """Test VAE anomaly detection functionality."""

    @classmethod
    def setUpClass(cls):
        """Set up test data once for all tests."""
        np.random.seed(42)
        dates = pd.date_range(start='2020-01-01', periods=200, freq='D')

        # Create normal data with patterns
        data = {
            'series_1': np.random.normal(100, 10, 200),
            'series_2': np.random.normal(50, 5, 200)
            + 10 * np.sin(np.arange(200) * 2 * np.pi / 7),
            'series_3': np.random.normal(75, 8, 200)
            + 5 * np.cos(np.arange(200) * 2 * np.pi / 30),
        }

        # Add some anomalies
        data['series_1'][50] = 200  # spike
        data['series_1'][51] = 210  # spike
        data['series_2'][100] = 0  # drop
        data['series_3'][150] = 150  # spike

        cls.df = pd.DataFrame(data, index=dates)

    def test_vae_availability(self):
        """Test that VAEOutlier is available in the methods."""
        print("Starting test_vae_availability")
        from autots.tools.anomaly_utils import available_methods

        self.assertIn(
            "VAEOutlier", available_methods, "VAEOutlier not found in available methods"
        )

    def test_vae_parameter_generation(self):
        """Test parameter generation for VAEOutlier."""
        print("Starting test_vae_parameter_generation")
        params = AnomalyDetector.get_new_params(method="VAEOutlier")
        self.assertEqual(params['method'], 'VAEOutlier', "Method should be VAEOutlier")
        self.assertIn('method_params', params, "method_params should be present")

        # Check that key parameters are present
        method_params = params['method_params']
        expected_params = [
            'depth',
            'batch_size',
            'epochs',
            'learning_rate',
            'loss_function',
            'dropout_rate',
            'beta',
            'contamination',
        ]
        for param in expected_params:
            self.assertIn(param, method_params, f"Parameter {param} should be present")

    def test_vae_anomaly_detection(self):
        """Test VAE anomaly detection functionality."""
        print("Starting test_vae_anomaly_detection")
        try:
            import torch

            torch_available = True
        except ImportError:
            self.skipTest("PyTorch not available, skipping VAE test")

        detector = AnomalyDetector(
            method="VAEOutlier",
            method_params={
                'depth': 1,
                'batch_size': 32,
                'epochs': 10,  # small for testing
                'learning_rate': 1e-3,
                'loss_function': 'elbo',
                'contamination': 0.1,
                'random_state': 42,
            },
        )

        anomalies, scores = detector.detect(self.df)

        # Check output shapes
        self.assertEqual(
            anomalies.shape, self.df.shape, "Anomalies shape should match input"
        )
        self.assertEqual(scores.shape, self.df.shape, "Scores shape should match input")

        # Check that some anomalies were detected
        num_anomalies = np.sum((anomalies == -1).values)
        self.assertGreater(num_anomalies, 0, "Should detect at least some anomalies")
        self.assertLess(
            num_anomalies, len(self.df) * 0.5, "Should not detect too many anomalies"
        )

        # Check that values are in expected range
        self.assertTrue(
            np.all(np.isin(anomalies.values, [-1, 1])), "Anomalies should be -1 or 1"
        )
        self.assertTrue(np.all(scores.values >= 0), "Scores should be non-negative")

    def test_vae_univariate_detection(self):
        """Test VAE anomaly detection in univariate mode."""
        print("Starting test_vae_univariate_detection")
        try:
            import torch

            torch_available = True
        except ImportError:
            self.skipTest("PyTorch not available, skipping VAE test")

        detector = AnomalyDetector(
            output='univariate',
            method="VAEOutlier",
            method_params={
                'depth': 1,
                'batch_size': 32,
                'epochs': 10,
                'learning_rate': 1e-3,
                'contamination': 0.1,
                'random_state': 42,
            },
        )

        anomalies, scores = detector.detect(self.df)

        # Check output shapes for univariate
        self.assertEqual(
            anomalies.shape,
            (len(self.df), 1),
            "Univariate anomalies should have single column",
        )
        self.assertEqual(
            scores.shape,
            (len(self.df), 1),
            "Univariate scores should have single column",
        )

    def test_vae_different_loss_functions(self):
        """Test VAE with different loss functions."""
        print("Starting test_vae_different_loss_functions")
        try:
            import torch

            torch_available = True
        except ImportError:
            self.skipTest("PyTorch not available, skipping VAE test")

        loss_functions = ['elbo', 'mse', 'lmse']

        for loss_func in loss_functions:
            with self.subTest(loss_function=loss_func):
                print(
                    f"Starting subtest test_vae_different_loss_functions loss_function={loss_func}"
                )
                detector = AnomalyDetector(
                    method="VAEOutlier",
                    method_params={
                        'depth': 1,
                        'batch_size': 32,
                        'epochs': 5,  # very small for testing
                        'learning_rate': 1e-3,
                        'loss_function': loss_func,
                        'contamination': 0.1,
                        'random_state': 42,
                    },
                )

                try:
                    anomalies, scores = detector.detect(self.df)
                    self.assertEqual(anomalies.shape, self.df.shape)
                    self.assertTrue(np.all(np.isin(anomalies.values, [-1, 1])))
                except Exception as e:
                    self.fail(f"VAE with {loss_func} loss failed: {e}")

    def test_vae_depth_parameter(self):
        """Test VAE with different depth parameters."""
        print("Starting test_vae_depth_parameter")
        try:
            import torch

            torch_available = True
        except ImportError:
            self.skipTest("PyTorch not available, skipping VAE test")

        for depth in [1, 2]:
            with self.subTest(depth=depth):
                print(f"Starting subtest test_vae_depth_parameter depth={depth}")
                detector = AnomalyDetector(
                    method="VAEOutlier",
                    method_params={
                        'depth': depth,
                        'batch_size': 32,
                        'epochs': 5,
                        'learning_rate': 1e-3,
                        'contamination': 0.1,
                        'random_state': 42,
                    },
                )

                try:
                    anomalies, scores = detector.detect(self.df)
                    self.assertEqual(anomalies.shape, self.df.shape)
                except Exception as e:
                    self.fail(f"VAE with depth {depth} failed: {e}")

    def test_vae_anomaly_removal_transformer(self):
        """Test AnomalyRemoval transformer with VAE method."""
        print("Starting test_vae_anomaly_removal_transformer")
        try:
            import torch

            torch_available = True
        except ImportError:
            self.skipTest("PyTorch not available, skipping VAE test")

        from autots.tools.transform import AnomalyRemoval

        transformer = AnomalyRemoval(
            method="VAEOutlier",
            method_params={
                'depth': 1,
                'batch_size': 32,
                'epochs': 5,
                'learning_rate': 1e-3,
                'contamination': 0.1,
                'random_state': 42,
            },
            fillna="mean",
        )

        try:
            cleaned_df = transformer.fit_transform(self.df)
            # Should have fewer or equal rows (anomalies removed)
            self.assertLessEqual(len(cleaned_df), len(self.df))
            # Should have same number of columns
            self.assertEqual(cleaned_df.shape[1], self.df.shape[1])
        except Exception as e:
            self.fail(f"AnomalyRemoval with VAE failed: {e}")


class TestAnomalyRemoval(unittest.TestCase):
    """Test AnomalyRemoval transformer functionality."""

    @classmethod
    def setUpClass(cls):
        """Set up test data once for all tests."""
        np.random.seed(42)
        dates = pd.date_range(start='2020-01-01', periods=200, freq='D')

        # Create normal data with patterns
        data = {
            'series_1': np.random.normal(100, 10, 200),
            'series_2': np.random.normal(50, 5, 200)
            + 10 * np.sin(np.arange(200) * 2 * np.pi / 7),
            'series_3': np.random.normal(75, 8, 200)
            + 5 * np.cos(np.arange(200) * 2 * np.pi / 30),
        }

        # Add some anomalies
        data['series_1'][50] = 200  # spike
        data['series_1'][51] = 210  # spike
        data['series_2'][100] = 0  # drop
        data['series_3'][150] = 150  # spike

        cls.df = pd.DataFrame(data, index=dates)

    def test_anomaly_removal_basic(self):
        """Test basic AnomalyRemoval functionality."""
        print("Starting test_anomaly_removal_basic")
        from autots.tools.transform import AnomalyRemoval

        transformer = AnomalyRemoval(
            method="zscore", method_params={"distribution": "norm", "alpha": 0.05}
        )

        result = transformer.fit_transform(self.df)

        # Should have fewer or equal rows (anomalies removed)
        self.assertLessEqual(len(result), len(self.df))
        # Should have same number of columns
        self.assertEqual(result.shape[1], self.df.shape[1])
        # Should have anomalies and scores attributes
        self.assertTrue(hasattr(transformer, 'anomalies'))
        self.assertTrue(hasattr(transformer, 'scores'))

    def test_anomaly_removal_with_fillna(self):
        """Test AnomalyRemoval with different fillna methods."""
        print("Starting test_anomaly_removal_with_fillna")
        from autots.tools.transform import AnomalyRemoval

        fillna_methods = ["ffill", "mean", "linear"]

        for fillna_method in fillna_methods:
            with self.subTest(fillna=fillna_method):
                print(
                    f"Starting subtest test_anomaly_removal_with_fillna fillna={fillna_method}"
                )
                transformer = AnomalyRemoval(
                    method="IQR",
                    method_params={"iqr_threshold": 2.0},
                    fillna=fillna_method,
                )

                result = transformer.fit_transform(self.df)

                # Should not have NaN values after fillna
                self.assertFalse(
                    result.isna().any().any(), f"NaN found with fillna={fillna_method}"
                )
                # Should have same shape or fewer rows
                self.assertLessEqual(len(result), len(self.df))
                self.assertEqual(result.shape[1], self.df.shape[1])

    def test_anomaly_removal_isolated_only(self):
        """Test AnomalyRemoval with isolated_only parameter."""
        print("Starting test_anomaly_removal_isolated_only")
        from autots.tools.transform import AnomalyRemoval

        # Without isolated_only
        transformer1 = AnomalyRemoval(
            method="zscore",
            method_params={"distribution": "norm", "alpha": 0.05},
            isolated_only=False,
        )
        result1 = transformer1.fit_transform(self.df)

        # With isolated_only
        transformer2 = AnomalyRemoval(
            method="zscore",
            method_params={"distribution": "norm", "alpha": 0.05},
            isolated_only=True,
        )
        result2 = transformer2.fit_transform(self.df)

        # isolated_only should generally remove fewer anomalies
        self.assertGreaterEqual(len(result2), len(result1))

    def test_anomaly_removal_multivariate_vs_univariate(self):
        """Test AnomalyRemoval with multivariate and univariate output."""
        print("Starting test_anomaly_removal_multivariate_vs_univariate")
        from autots.tools.transform import AnomalyRemoval

        # Multivariate
        transformer_multi = AnomalyRemoval(output='multivariate', method="IQR")
        result_multi = transformer_multi.fit_transform(self.df)

        # Univariate
        transformer_uni = AnomalyRemoval(output='univariate', method="IQR")
        result_uni = transformer_uni.fit_transform(self.df)

        # Both should return valid results
        self.assertLessEqual(len(result_multi), len(self.df))
        self.assertLessEqual(len(result_uni), len(self.df))
        self.assertEqual(result_multi.shape[1], self.df.shape[1])
        self.assertEqual(result_uni.shape[1], self.df.shape[1])

    def test_anomaly_removal_inverse_transform(self):
        """Test AnomalyRemoval inverse_transform."""
        print("Starting test_anomaly_removal_inverse_transform")
        from autots.tools.transform import AnomalyRemoval

        # Test with on_inverse=False (default)
        transformer1 = AnomalyRemoval(method="zscore", on_inverse=False)
        transformer1.fit(self.df)
        result1 = transformer1.inverse_transform(self.df)

        # Should return unchanged data
        pd.testing.assert_frame_equal(result1, self.df)

        # Test with on_inverse=True
        transformer2 = AnomalyRemoval(method="zscore", on_inverse=True)
        transformer2.fit(self.df)
        result2 = transformer2.inverse_transform(self.df)

        # Should apply fit_transform on inverse
        self.assertLessEqual(len(result2), len(self.df))

    def test_anomaly_removal_fit_anomaly_classifier(self):
        """Test fit_anomaly_classifier and score_to_anomaly methods."""
        print("Starting test_anomaly_removal_fit_anomaly_classifier")
        from autots.tools.transform import AnomalyRemoval

        transformer = AnomalyRemoval(method="IQR")
        transformer.fit(self.df)

        # Fit classifier
        transformer.fit_anomaly_classifier()

        # Should have classifier and categories
        self.assertIsNotNone(transformer.anomaly_classifier)
        self.assertIsNotNone(transformer.score_categories)

        # Test score_to_anomaly
        new_scores = transformer.scores.copy()
        anomaly_pred = transformer.score_to_anomaly(new_scores)

        # Should return valid anomaly classifications
        self.assertEqual(anomaly_pred.shape, new_scores.shape)
        self.assertTrue(all(np.isin(anomaly_pred.values.flatten(), [-1, 1])))

    def test_anomaly_removal_get_new_params(self):
        """Test get_new_params static method."""
        print("Starting test_anomaly_removal_get_new_params")
        from autots.tools.transform import AnomalyRemoval

        # Test different methods
        for method in ["random", "fast"]:
            with self.subTest(method=method):
                print(
                    f"Starting subtest test_anomaly_removal_get_new_params method={method}"
                )
                params = AnomalyRemoval.get_new_params(method=method)

                # Should have required keys
                self.assertIn('method', params)
                self.assertIn('method_params', params)
                self.assertIn('fillna', params)
                self.assertIn('isolated_only', params)
                self.assertIn('on_inverse', params)

    def test_anomaly_removal_with_transform_dict(self):
        """Test AnomalyRemoval with transform_dict preprocessing."""
        print("Starting test_anomaly_removal_with_transform_dict")
        from autots.tools.transform import AnomalyRemoval

        transformer = AnomalyRemoval(
            method="zscore",
            transform_dict={
                "transformations": {0: "ClipOutliers"},
                "transformation_params": {0: {"method": "clip", "std_threshold": 3}},
            },
        )

        result = transformer.fit_transform(self.df)

        # Should complete without error
        self.assertLessEqual(len(result), len(self.df))
        self.assertEqual(result.shape[1], self.df.shape[1])

    def test_anomaly_removal_multiple_methods(self):
        """Test AnomalyRemoval with different anomaly detection methods."""
        print("Starting test_anomaly_removal_multiple_methods")
        from autots.tools.transform import AnomalyRemoval
        from autots.tools.anomaly_utils import fast_methods

        for method in fast_methods[:3]:  # Test first 3 fast methods
            with self.subTest(method=method):
                print(
                    f"Starting subtest test_anomaly_removal_multiple_methods method={method}"
                )
                try:
                    transformer = AnomalyRemoval(method=method)
                    result = transformer.fit_transform(self.df)

                    # Should complete without error
                    self.assertLessEqual(len(result), len(self.df))
                    self.assertEqual(result.shape[1], self.df.shape[1])
                except Exception as e:
                    # Some methods might fail on this small dataset, that's okay
                    if "requires" not in str(e).lower():
                        raise


if __name__ == '__main__':
    unittest.main()
